# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2024, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from .utils import piping
from ._unit import Unit
from ._facility import Facility

__all__ = ('Module', 'FacilityModule')

class Module(Unit):
    _ins_size_is_fixed = _outs_size_is_fixed = False
    _N_ins = _N_outs = 0
    
    def _init(self, system, run_full_simulation=None):
        if not self._ins: self._ins = piping.StreamPorts.from_inlets(system.feeds)
        if not self._outs: self._outs = piping.StreamPorts.from_outlets(system.products)
        self.register_auxiliary(system, 'auxsystem')
        if run_full_simulation is None: run_full_simulation = bool(system.facilities)
        self.run_full_simulation = run_full_simulation
    
    def _assert_compatible_property_package(self): pass
    
    def _setup(self):
        self.auxsystem._setup(load_configuration=False)
        super()._setup()    
    
    def _run(self):
        if self.run_full_simulation:
            self.auxsystem.simulate()
        else:
            self.auxsystem.converge()

        
class FacilityModule(Module, Facility): network_priority = None


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# Copyright (C) 2022, Yoel Cortes-Pena <yoelcortes@gmail.com> and Ben Portner <https://github.com/BenPortner>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import numpy as np
from biosteam.utils.piping import (
    ignore_docking_warnings, Connection, 
)
from warnings import warn
import biosteam as bst
from graphviz import Digraph
from IPython import display
from thermosteam import AbstractStream, AbstractUnit
from xml.etree import ElementTree
from typing import Optional
import urllib
import re

__all__ = ('digraph_from_system',
           'digraph_from_units',
           'digraph_from_units_and_connections',
           'update_digraph_from_units_and_connections',
           'blank_digraph',
           'minimal_digraph',
           'surface_digraph',
           'finalize_digraph',
           'extend_surface_units',
           'display_digraph',
           'save_digraph')

stream_node = dict(
    fillcolor='#55a8b5',
    fontcolor='white', 
    style='filled', 
    orientation='0',
    width='0.6', 
    height='0.6', 
    color='#90918e', 
    margin='0',
    peripheries='1',
    fontname="Arial",
)

class PenWidth:
    __slots__ = ('name', 'percentiles')
    def __init__(self, name, streams):
        self.name = name
        values = [s.get_property(name) for s in streams if not s.isempty()] or [0] 
        try:
            self.percentiles = np.percentile(
                values, 
                [33, 66, 100]
            )
        except:
            self.percentiles = 3 * [max(values)]
    
    def __call__(self, stream):
        value = stream.get_property(self.name)
        for width, percentile in enumerate(self.percentiles, 1):
            if value <= percentile: return str(width * 0.6 + 0.4)
        return str(width * 0.6 + 0.4)
    
    def __repr__(self):
        return f"{type(self).__name__}(name={repr(self.name)}, scale={self.scale:.3g})"


preferences = bst.preferences

def sort_streams(streams):
    return sorted(streams, key=lambda x: -x.F_mass + len(x.ID))

def has_path(obj):
    return hasattr(obj, 'path')

def blank_digraph(format='svg', maxiter='10000000000000000000', 
                  Damping='0.995', K='0.5', **graph_attrs):
    # Create a digraph and set direction left to right
    f = Digraph(format=format)
    f.attr(rankdir='LR', maxiter=maxiter, Damping=Damping, K=K, 
           penwidth='0', color='none', bgcolor=preferences.background_color,
           fontcolor=preferences.label_color, fontname="Arial",
           labeljust='l', labelloc='t', fontsize='24', constraint='false',
           **graph_attrs)
    return f

def get_section_inlets_and_outlets(units, streams):
    outs = []
    ins = []
    for s in streams:
        source = s._source
        sink = s._sink
        if source in units and sink not in units:
            outs.append(s)
        elif sink in units and source not in units:
            ins.append(s)
    return ins, outs

@ignore_docking_warnings
def minimal_digraph(ID, units, streams, auxiliaries=None, **graph_attrs):
    ins, outs = get_section_inlets_and_outlets(units, streams)
    thermo_water = bst.HeatUtility.thermo_water
    product = bst.Stream('.', thermo=thermo_water, Water=sum([i.F_mass for i in outs]), units='kg/hr')
    feed = bst.Stream('.', thermo=thermo_water, Water=sum([i.F_mass for i in ins]), units='kg/hr')
    ins = sort_streams(ins)
    outs = sort_streams(outs)
    system_box = bst.units.DiagramOnlySystemUnit(ID, feed, product)
    boxes = [system_box]
    if preferences.show_all_streams:
        name = '\n'.join([i.ID for i in ins])
    else:
        name = '\n'.join([i.ID for i in ins if not i.isempty()])
    if name:
        feed_box = bst.units.DiagramOnlyStreamUnit(name, None, feed)
        boxes.append(feed_box)
    if preferences.show_all_streams:
        name = '\n'.join([i.ID for i in outs])
    else:
        name = '\n'.join([i.ID for i in outs if not i.isempty()])
    if name:
        product_box = bst.units.DiagramOnlyStreamUnit(name, product, None)
        boxes.append(product_box)
    return digraph_from_units(boxes, auxiliaries=False,
                              **graph_attrs)

@ignore_docking_warnings
def surface_digraph(path, auxiliaries=None, **graph_attrs):
    surface_units = []
    old_unit_connections = set()
    isa = isinstance
    Unit = AbstractUnit
    done = set()
    for i in path:
        if i in done: continue
        done.add(i)
        if isa(i, Unit):
            surface_units.append(i)
        elif has_path(i):
            extend_surface_units(i.ID, i.streams, i.units, 
                                 surface_units, old_unit_connections)
    f = digraph_from_units(surface_units, auxiliaries=False, **graph_attrs)
    for u, ins, outs in old_unit_connections:
        u._ins[:] = ins
        u._outs[:] = outs    
    return f

@ignore_docking_warnings
def extend_surface_units(ID, streams, units, surface_units, old_unit_connections):
    outs = []
    ins = []
    feeds = []
    products = []
    StreamUnit = bst.units.DiagramOnlyStreamUnit
    SystemUnit = bst.units.DiagramOnlySystemUnit
    
    for s in streams:
        source = s._source
        sink = s._sink
        if source in units and sink not in units:
            if sink: outs.append(s)
            else: products.append(s)
            u_io = (source, tuple(source.ins), tuple(source.outs))
            old_unit_connections.add(u_io)
        elif sink in units and source not in units:
            if source: ins.append(s)
            else: feeds.append(s)
            u_io = (sink, tuple(sink.ins), tuple(sink.outs))
            old_unit_connections.add(u_io)
    
    if len(feeds) > 1:
        thermo_water = bst.HeatUtility.thermo_water
        feed = bst.Stream('.', thermo=thermo_water, Water=sum([i.F_mass for i in feeds]), units='kg/hr')
        feeds = sort_streams(feeds)
        if preferences.show_all_streams:
            name = '\n'.join([i.ID for i in feeds])
        else:
            name = '\n'.join([i.ID for i in feeds if not i.isempty()])
        if name:
            feed_box = bst.units.DiagramOnlyStreamUnit(name.replace('_', ' '), None, feed)
            ins.append(feed)
        else:
            feed_box = None
    else: 
        feed_box = None
        ins += feeds
    
    if len(products) > 1:
        thermo_water = bst.HeatUtility.thermo_water
        product = bst.Stream('.', thermo=thermo_water, Water=sum([i.F_mass for i in products]), units='kg/hr')
        products = sort_streams(products)
        if preferences.show_all_streams:
            name = '\n'.join([i.ID for i in products])
        else:
            name = '\n'.join([i.ID for i in products if not i.isempty()])
        if name:
            product_box = bst.units.DiagramOnlyStreamUnit(name.replace('_', ' '), product, None)
            outs.append(product)
        else:
            product_box = None
    else: 
        product_box = None
        outs += products
    
    subsystem_unit = SystemUnit(ID, ins, outs)
    for i in (feed_box, subsystem_unit, product_box):
        if i: surface_units.append(i)

def digraph_from_units(units, streams=None, auxiliaries=None, **graph_attrs):
    if auxiliaries:
        all_units = []
        streams = []
        stream_set = set()
        for unit in tuple(units):
            all_units.append(unit)
            for s in unit.ins + unit.outs:
                if s in stream_set: continue
                streams.append(s)
                stream_set.add(s)
            for name, auxunit in unit._diagram_nested_auxiliary_units_with_names(depth=auxiliaries):
                auxunit.owner = unit # In case units not created correctly through Unit.auxiliary method
                if isinstance(auxunit, AbstractUnit): 
                    auxunit._ID = name
                    all_units.append(auxunit)
                    for s in auxunit.ins + auxunit.outs:
                        if s in stream_set: continue
                        streams.append(s)
                        stream_set.add(s)
        units = all_units
    else:
        if streams is None:
            streams = []
            stream_set = set()
            for u in units:
                for s in u._ins + u._outs:
                    if s in stream_set: continue
                    streams.append(s)
                    stream_set.add(s)
        else:
            stream_set = set(streams)
    if auxiliaries:
        for stream in tuple(streams):
            if hasattr(stream, 'port'):
                other = stream.port.get_stream()
                if other in stream_set:
                    streams.remove(other)
                    stream_set.remove(other)
    digraph = digraph_from_units_and_connections(units, get_all_connections(streams), auxiliaries, **graph_attrs)
    return digraph

def digraph_from_system(system, path=None, auxiliaries=None, **graph_attrs):
    f = blank_digraph(**graph_attrs) 
    other_streams = set()
    excluded_connections = set()
    if not path: path = system.path
    unit_names = get_unit_names(f, path, with_auxiliaries=False)
    update_digraph_from_path(f, path, 
                             system.recycle, 0, unit_names, excluded_connections,
                             other_streams)
    connections = get_all_connections(other_streams, excluded_connections)
    add_connections(f, connections, unit_names)
    return f


def update_digraph_from_path(f, path, recycle, depth, unit_names,
                             excluded_connections,
                             other_streams):
    all_streams = set()
    units = set()
    subsystems = []
    isa = isinstance
    System = bst.System
    Unit = AbstractUnit
    for i in path:
        if isa(i, Unit):
            units.add(i)
            all_streams.update(i._ins + i._outs)
        elif isa(i, System): 
            subsystems.append(i)
    if isa(recycle, AbstractStream): 
        recycles = [recycle] 
    elif recycle:
        recycles = recycle
    else:
        recycles = []
    streams = [i for i in all_streams if (not i.sink or i.sink in units) and (not i.source or i.source in units)]
    other_streams.update(all_streams.difference(streams))
    connections = get_all_connections(recycles, excluded_connections)
    add_connections(f, connections, unit_names, color='#f1777f', fontcolor='#f1777f')
    connections = get_all_connections(streams, excluded_connections)
    add_connections(f, connections, unit_names)
    depth += 1
    N_colors = len(preferences.depth_colors)
    color = preferences.depth_colors[(depth - 1) % N_colors]
    if preferences.fill_cluster:
        kwargs = dict(bgcolor=color, penwidth='0.2', color=preferences.stream_color)
    else:
        kwargs = dict(color=color, bgcolor='none', penwidth='0.75', style='solid')
    for i in subsystems:
        with f.subgraph(name='cluster_' + i.ID) as c:
            c.attr(label=i.ID, fontname="Arial", 
                   labeljust='l', fontcolor=preferences.label_color, **kwargs)
            update_digraph_from_path(c, i.path, i.recycle, depth, unit_names, excluded_connections, other_streams)

def digraph_from_units_and_connections(units, connections, with_auxiliaries, **graph_attrs):
    f = blank_digraph(**graph_attrs)
    update_digraph_from_units_and_connections(f, units, connections, with_auxiliaries)
    return f

def fill_info_from_path(path, indices, info_by_unit):
    isa = isinstance
    number = preferences.number_path
    profile = preferences.profile
    Timer = bst.utils.Timer
    for u in path:
        if isa(u, bst.Junction):
            continue # Do no include junctions
        if isa(u, bst.System):
            fill_info_from_path(u.path, [*indices, 0], info_by_unit)
            indices[-1] += 1
            continue
        index = '.'.join([str(i) for i in indices])
        indices[-1] += 1
        if u in info_by_unit:
            old_data = info_by_unit[u]
            if number: old_data[0].append(index)
        else:
            if profile: # pragma: no cover
                t = Timer()
                for n in range(1):
                    t.start(); u.simulate(); t.elapsed_time()
                    if n > 1 and sum(t.record) > 0.2: break 
                time = f"{1000 * t.mean:.2g} ms"
            else:
                time = None
            info_by_unit[u] = [[index] if number else [], time] 

def get_unit_names(f: Digraph, path, with_auxiliaries):
    unit_names = {}  # Contains full description (ID and line) by unit
    info_by_unit = {}
    fill_info_from_path(path, [0], info_by_unit)
    for u, (index, time) in info_by_unit.items():
        node = u.get_node()
        if 'xlabel' in node:
            key = 'xlabel'
        else:
            key = 'label'
        label = node[key]
        info = ', '.join(index) 
        if time is not None:
            if info:
                info = f"{info}; {time}"
            else:
                info = time
        if info: label = f"[{info}] {label}"
        node[key] = label
        unit_names[u] = node['name']
        if (with_auxiliaries and all([i in info_by_unit for i in u.auxiliary_units])
            and u._assembled_from_auxiliary_units()): continue
        f.node(**node)
    return unit_names

def update_digraph_from_units_and_connections(f: Digraph, units, connections, with_auxiliaries):
    add_connections(f, connections, get_unit_names(f, units, with_auxiliaries))    

def get_all_connections(streams, added_connections=None):
    if added_connections is None: added_connections = set()
    connections = []
    originals = {}
    superinlet = lambda s: s.__class__.__name__ == 'SuperpositionInlet'
    superoutlet = lambda s: s.__class__.__name__ == 'SuperpositionOutlet'
    for s in streams:
        original = s
        while hasattr(original, 'port'):
            original = original.port.get_stream()
        if (s._source or s._sink): 
            connection = s.get_connection(junction=False)
            if original in originals:
                source0, source_index0, stream0, sink_index0, sink0 = old_connection = originals[original]
                source1, source_index1, stream1, sink_index1, sink1 = connection
                if superoutlet(stream1) and superinlet(stream0):
                    connections.remove(old_connection)
                    added_connections.remove(old_connection)
                    stream = stream1.copy(ID='.' + s.ID)
                    connection = Connection(source1, source_index1, stream, sink_index0, sink0)
                elif superoutlet(stream0) and superinlet(stream1):
                    connections.remove(old_connection)
                    added_connections.remove(old_connection)
                    stream = stream1.copy(ID='.' + s.ID)
                    connection = Connection(source0, source_index0, stream, sink_index1, sink1)
            originals[original] = connection
            if connection and connection not in added_connections:
                connections.append(connection)
                added_connections.add(connection)
    return connections

def add_connection(f: Digraph, connection, unit_names, pen_width=None, **edge_options):
    source, source_index, stream, sink_index, sink = connection
    has_source = source in unit_names
    has_sink = sink in unit_names
    style = 'dashed' if (
        stream.isempty()
        and not isinstance(stream.source, bst.units.DiagramOnlyStreamUnit)
        and not isinstance(stream.sink, bst.units.DiagramOnlyStreamUnit)
    ) else 'solid'
    f.attr('edge', label='', taillabel='', headlabel='', labeldistance='2',
           **edge_options)
    tooltip = stream._get_tooltip_string(bst.preferences.graphviz_format, bst.preferences.tooltips_full_results)
    if stream:
        lines = []
        line = ''
        ID = stream.ID
        if ID:
            for word in ID.split('_'):
                line += ' ' + word
                if len(line) > 10: 
                    lines.append(line)
                    line = ''
            if line: lines.append(line)
            ID = '\n'.join(lines)
        else:
            ID = ' '
        ref = str(hash(stream))
        penwidth = pen_width(stream) if pen_width else '1.0'
        # Make stream nodes / unit-stream edges / unit-unit edges
        if has_sink and not has_source:
            # Feed stream case
            if not preferences.show_all_streams and stream.isempty(): return
            f.node(ref,
                   width='0.15', 
                   height='0.15',
                   shape='diamond',
                   fillcolor='#f98f60',
                   color=preferences.stream_color,
                   label='')
            inlet_options = sink._graphics.get_inlet_options(sink, sink_index)
            f.attr('edge', arrowtail='none', arrowhead='none', label=ID,
                   tailport='e', style=style, penwidth=penwidth, **inlet_options)
            f.edge(ref, unit_names[sink], labeltooltip=tooltip)
        elif has_source and not has_sink:
            # Product stream case
            if not preferences.show_all_streams and stream.isempty(): return
            f.node(ref, 
                   width='0.15', 
                   height='0.2',
                   shape='triangle',
                   orientation='270',
                   fillcolor='#f98f60',
                   color=preferences.stream_color,
                   label='')
            outlet_options = source._graphics.get_outlet_options(source, source_index)
            f.attr('edge', arrowtail='none', arrowhead='none', label=ID,
                   headport='w', style=style, penwidth=penwidth, **outlet_options)
            f.edge(unit_names[source], ref, labeltooltip=tooltip)
        elif has_sink and has_source:
            # Process stream case
            inlet_options = sink._graphics.get_inlet_options(sink, sink_index)
            outlet_options = source._graphics.get_outlet_options(source, source_index)
            f.attr('edge', arrowtail='none', arrowhead='normal', style=style, 
                   **inlet_options, penwidth=penwidth, **outlet_options)
            label = ID if preferences.label_streams else ' '
            f.edge(unit_names[source], unit_names[sink], label=label, labeltooltip=tooltip)
    elif has_sink and has_source:
        # Missing process stream case
        inlet_options = sink._graphics.get_inlet_options(sink, sink_index)
        outlet_options = source._graphics.get_outlet_options(source, source_index)
        f.attr('edge', arrowtail='none', arrowhead='normal', label=' ',
               **inlet_options, **outlet_options)
        f.edge(unit_names[source], unit_names[sink], style='dashed', labeltooltip=tooltip)

def add_connections(f: Digraph, connections, unit_names, color=None, fontcolor=None, **edge_options):
    stream_width = preferences.stream_width
    if stream_width:
        pen_width = PenWidth(stream_width, [i.stream for i in connections])
    else:
        pen_width = None
    
    # Set attributes for graph and streams
    f.attr('graph', overlap='orthoyx', fontname="Arial", outputorder='edgesfirst', 
           nodesep='0.5', ranksep='0.15', maxiter='1000000')
    f.attr('edge', dir='foward', fontname='Arial')
    f.attr('node', **stream_node)
    index = {j: i for i, j in unit_names.items()}
    length = len(index)
    def key(x):
        value = index.get(x.source, 0) + index.get(x.sink, length)
        if x.source_index:
            value += 1e-3 * x.source_index / len(x.source.outs)
        if x.sink_index:
            value += 1e-6 * x.sink_index / len(x.sink.ins)
        return value
    connections = sorted(connections, key=key)
    for connection in connections:
        add_connection(f, connection, unit_names, 
                       color=color or preferences.stream_color,
                       fontcolor=fontcolor or preferences.label_color,
                       pen_width=pen_width,
                       **edge_options)

def fix_valve_symbol_in_svg_output(
        img:bytes,
        unit_color: Optional[str] = None,
        unit_periphery_color: Optional[str] = None,
        label_color: Optional[str] = None,
):
    """Fix valve symbols because images cannot be loaded when choosing `format='svg'`"""
    if unit_color is None: unit_color = bst.preferences.unit_color
    if unit_periphery_color is None: unit_periphery_color = bst.preferences.unit_periphery_color
    if label_color is None: label_color = bst.preferences.unit_label_color
    # get all image tags
    tree = ElementTree.fromstring(img)
    images = [e for e in tree.iter() if 'image' in e.tag]
    # make polygon
    parent_map = {c: p for p in tree.iter() for c in p}
    getchildren = lambda pm: pm.getchildren() if hasattr(pm, 'getchildren') else list(pm)
    polygons = [c for i in images for c in getchildren(parent_map[i]) if 'polygon' in c.tag]
    for p in polygons:
        points = p.attrib["points"].split(' ')
        # turn rect into valve symbol
        buffer = points[1]
        points[1] = points[2]
        points[2] = buffer
        p.attrib["points"] = ' '.join(points)
        # fix color
        p.attrib["fill"] = unit_color.split(':')[-1] # In case of gradiant color (white:#CDCDCD)
        p.attrib["stroke"] = unit_periphery_color
    # fix label text color and position
    label_image = [(c, i) for i in images for c in getchildren(parent_map[parent_map[parent_map[i]]]) if 'text' in c.tag]
    for l,i in label_image:
        l.attrib["fill"] = label_color
        width = float(i.attrib['width'].replace('px',''))
        x = float(i.attrib["x"])
        l.attrib["x"] = f"{x+width/2}"
    # delete image tags
    for i in images:
        parent_map[i].remove(i)
    return ElementTree.tostring(tree)

def inject_javascript(img:bytes):
    html = ElementTree.Element('html')
    head = ElementTree.SubElement(html, 'head')
    # insert css
    links = [
        "https://unpkg.com/tippy.js@6.3.7/themes/translucent.css",
        "https://rawcdn.githack.com/BioSTEAMDevelopmentGroup/biosteam/e065aca079c216d72b75949bbcbb74a3bbddb75d/biosteam/digraph/digraph.css",
    ]
    for href in links:
        link = ElementTree.SubElement(head, 'link')
        link.set("rel", "stylesheet")
        link.set("href", href)
    # insert javascript
    srcs = [
        "https://unpkg.com/@popperjs/core@2",
        "https://unpkg.com/tippy.js@6",
        "https://rawcdn.githack.com/BioSTEAMDevelopmentGroup/biosteam/e065aca079c216d72b75949bbcbb74a3bbddb75d/biosteam/digraph/digraph.js",
    ]
    for src in srcs:
        script = ElementTree.SubElement(head, 'script')
        script.set("src", src)
    # body
    body = ElementTree.SubElement(html, 'body')
    svg = ElementTree.fromstring(img)
    
    # getiterator is deprecated in Python 3.9
    getiter = lambda etree: (getattr(etree, 'getiterator', None) or getattr(etree, 'iter'))()
    
    # remove namespaces from tags and attributes
    for e in getiter(svg):
        e.tag = re.sub("{.*?}","",e.tag)
        for key, value in e.attrib.copy().items():
            if "{" in key:
                clean = re.sub("{.*?}","",key)
                e.attrib[clean] = value
                del e.attrib[key]
    # make tippy tooltips
    for e in getiter(svg):
        for key, value in e.attrib.copy().items():
            if key == "class" and value in ["node", "edge"]:
                title = e.find("./title")
                if title is not None:
                    default_tooltip = title.text
                else:
                    default_tooltip = e.text
                custom_tooltip = (e.find("./g/a") or e).attrib.get("title", None)
                tooltip = custom_tooltip or default_tooltip
                if tooltip is not None:
                    e.attrib["data-tippy-content"] = tooltip.strip()
    # remove default tooltips
    for e in getiter(svg):
        t = e.find("./title")
        if t is not None:
            e.remove(t)
    body.append(svg)
    # add docstring declaration
    s = ElementTree.tostring(html, encoding='utf8', method='html')
    s = b"<!DOCTYPE html>"+s
    return s

def display_digraph(digraph, format, height=None): # pragma: no coverage
    if format is None: format = preferences.graphviz_format
    if height is None: height = '400px'
    if format == 'svg':
        img = digraph.pipe(format='svg')
        img = fix_valve_symbol_in_svg_output(img)
        x = display.SVG(img)
        display.display(x)
    # TODO: Consult about this complicated Javascript injection
    elif format == 'html':
        img = digraph.pipe(format='svg')
        img = fix_valve_symbol_in_svg_output(img)
        img = inject_javascript(img)
        data_uri = 'data:text/html;charset=utf-8,' + urllib.parse.quote(img)
        x = display.IFrame(src=data_uri, width='100%', height=height,
                            extras=['allowtransparency="true"'])
        display.display(x)
    else:
        x = display.Image(digraph.pipe(format='png'))
        display.display(x)

def save_digraph(digraph, file, format): # pragma: no coverage
    if '.' not in file:
        if format is None: format = preferences.graphviz_format
        file += '.' + format
    elif format is None:
        format = file.split()[-1]
    else:
        raise ValueError(
            "cannot specify format extension; file already has format "
           f"extension '{file.split()[-1]}'"
        )
    if format == 'html':
        try:
            img = digraph.pipe(format='svg')
        except Exception as e:
            try:
                from signal import signal, SIGPIPE, SIG_DFL
                signal(SIGPIPE, SIG_DFL)
            except ImportError:
                raise e
        img = fix_valve_symbol_in_svg_output(img)
        img = inject_javascript(img)
    else:
        img = digraph.pipe(format=format)
        if format == 'svg': img = fix_valve_symbol_in_svg_output(img)
    f = open(file, 'wb')
    f.write(img)
    f.close()
    
def finalize_digraph(digraph, file, format, height=None): # pragma: no coverage
    if preferences.raise_exception: 
        if file: save_digraph(digraph, file, format)
        else: display_digraph(digraph, format, height)
    else:
        try:
            if file: save_digraph(digraph, file, format)
            else: display_digraph(digraph, format, height)
        except (OSError, TypeError) as exp:
            raise exp from None
        except Exception as exp: 
            warn(
                f"a '{type(exp).__name__}' was raised when generating "
                "graphviz diagram, possibly due to graphviz installation issues, "
                "make sure Graphviz executables are on your systems' PATH",
                RuntimeWarning
            )


# -*- coding: utf-8 -*-
"""
"""
import pytest
import biosteam as bst
import thermosteam as tmo
from math import exp, log
import numpy as np
from thermosteam.constants import R
from numpy.testing import assert_allclose

def test_multi_stage_adiabatic_vle():
    bst.settings.set_thermo(['AceticAcid', 'EthylAcetate', 'Water', 'MTBE'], cache=True)
    feed = bst.Stream('feed', Water=75, AceticAcid=5, MTBE=20, T=320)
    steam = bst.Stream('steam', Water=100, phase='g', T=390)
    MSE = bst.MultiStageEquilibrium(N_stages=2, ins=[feed, steam], feed_stages=[0, -1],
        outs=['vapor', 'liquid'],
        phases=('g', 'l'),
    )
    MSE.simulate()
    with bst.System() as sys:
        vapor_recycle = bst.Stream()
        M0 = bst.Mixer(ins=[feed.copy(), vapor_recycle], conserve_phases=True)
        stage0 = bst.Flash(ins=M0-0, P=101325, Q=0)
        M1 = bst.Mixer(ins=[stage0.liquid, steam.copy()], conserve_phases=True)
        stage1 = bst.Flash(ins=M1-0,
                           outs=[vapor_recycle, 'liquid'], 
                           P=101325, Q=0)
    sys.set_tolerance(mol=0.01, rmol=0.001)
    sys.simulate()
    assert_allclose(
        MSE.stages[0].vapor.mol, 
        stage0.vapor.mol,
        atol=1,
        rtol=0.01,
    )
    assert_allclose(
        MSE.stages[0].liquid.mol, 
        stage0.liquid.mol,
        atol=1,
        rtol=0.01,
    )
    assert_allclose(
        MSE.stages[1].vapor.mol, 
        stage1.vapor.mol,
        atol=1,
        rtol=0.01,
    )
    assert_allclose(
        MSE.stages[1].liquid.mol, 
        stage1.liquid.mol,
        atol=1,
        rtol=0.01,
    )
    assert_allclose(
        MSE.stages[1].vapor.T, 
        stage1.vapor.T,
        atol=1,
        rtol=0.01,
    )
    assert_allclose(
        MSE.stages[1].liquid.T, 
        stage1.liquid.T,
        atol=1,
        rtol=0.01,
    )
    
def test_multi_stage_adiabatic_vle_non_condensables():
    import biosteam as bst
    bst.settings.set_thermo(['Water', 'Ethanol', 'CO2', 'N2'], cache=True)
    liq = bst.Stream('feed', Water=100, units='kg/hr', T=320)
    gas = bst.Stream('steam', CO2=100, N2=100, Ethanol=40, units='kg/hr', phase='g', T=305.15)
    MSE = bst.MultiStageEquilibrium(N_stages=10, ins=[liq, gas], feed_stages=[0, -1],
        outs=['vapor', 'liquid'],
        phases=('g', 'l'),
    )
    MSE.simulate()

def test_small_esterification_column():
    import biosteam as bst
    bst.settings.set_thermo(['Water', 'LacticAcid', 'MethylLactate', 'Methanol', 'SuccinicAcid'], cache=True)
    from math import exp, log
    from thermosteam.constants import R
    
    class Esterification(bst.KineticReaction):
        catalyst_fraction = 0.5 
        
        def volume(self, stream): # kg of catalyst
            rho_cat = 770 # kg / m3
            liquid_volume = self.liquid_volume
            catalyst_volume = self.catalyst_fraction * liquid_volume
            catalyst_mass = catalyst_volume * rho_cat
            return catalyst_mass
        
        def rate(self, stream):
            T = stream.T
            kf = 2.98e4 * exp(-51300 / (R * T))
            kr = 2.74e2 * exp(-47200 / (R * T))
            H2O, La, LaMe, MeOH, _ = stream.mol / stream.F_mol
            return 3600 * (kf * La * MeOH - kr * LaMe * H2O) # kmol / kg-catalyst / hr
    
    
    feed = bst.Stream(
        'feed',
        LacticAcid=4.174,
        Water=5.460,
        SuccinicAcid=0.531,
        MethylLactate=1e-6,
        total_flow=10.165,
        P=0.106 * 101325,
        T=72.5 + 273.15
    )
    feed.T = feed.bubble_point_at_P(P=0.2 * 101325).T
    methanol = bst.Stream('methanol', Methanol=16.695, P=0.2 * 101325, phase='g')
    methanol.T = methanol.dew_point_at_P().T
    esterification = bst.MESHDistillation(
        'esterification',
        ins=(feed, methanol), 
        outs=('esterification_distillate', 'bottoms'),
        N_stages=3,
        feed_stages=(1, 1),
        reflux=1, # Unknown (update later)
        boilup=1,
        # bottoms_product_to_feed=0.0334, # Fraction of feed.
        full_condenser=True,
        stage_reactions={
            i: Esterification('LacticAcid + Methanol -> Water + MethylLactate', reactant='LacticAcid')
            for i in range(1, 2)
        },
        maxiter=50,
        LHK=('Methanol', 'MethylLactate'),
        P=0.2 * 101325,
        use_cache=True,
        algorithms=('inside out',)
    )
    esterification.simulate()

def test_esterification_column():
    import biosteam as bst
    bst.settings.set_thermo(['Water', 'LacticAcid', 'MethylLactate', 'Methanol', 'SuccinicAcid'], cache=True)
    from math import exp, log
    from thermosteam.constants import R
    
    class Esterification(bst.KineticReaction):
        catalyst_fraction = 0.5 
        
        def volume(self, stream): # kg of catalyst
            rho_cat = 770 # kg / m3
            liquid_volume = self.liquid_volume
            catalyst_volume = self.catalyst_fraction * liquid_volume
            catalyst_mass = catalyst_volume * rho_cat
            return catalyst_mass
        
        def rate(self, stream):
            T = stream.T
            kf = 2.98e4 * exp(-51300 / (R * T))
            kr = 2.74e2 * exp(-47200 / (R * T))
            H2O, La, LaMe, MeOH, _ = stream.mol / stream.F_mol
            rate = 3600 * (kf * La * MeOH - kr * LaMe * H2O) # kmol / kg-catalyst / hr
            print(rate)
            return 0
    
    
    feed = bst.Stream(
        'feed',
        LacticAcid=4.174,
        Water=5.460,
        SuccinicAcid=0.531,
        MethylLactate=1e-6,
        total_flow=10.165,
        P=0.106 * 101325,
        T=72.5 + 273.15
    )
    feed.T = feed.bubble_point_at_P(P=0.2 * 101325).T
    methanol = bst.Stream('methanol', Methanol=16.695, P=0.2 * 101325, phase='g')
    methanol.T = methanol.dew_point_at_P().T
    esterification = bst.MESHDistillation(
        'esterification',
        ins=(feed, methanol), 
        outs=('esterification_distillate', 'bottoms'),
        N_stages=21,
        feed_stages=(1, 19),
        reflux=1, # Unknown (update later)
        boilup=1,
        # bottoms_product_to_feed=0.0334, # Fraction of feed.
        full_condenser=True,
        stage_reactions={
            i: Esterification('LacticAcid + Methanol -> Water + MethylLactate', reactant='LacticAcid')
            for i in range(1, 19)
        },
        maxiter=50,
        LHK=('Methanol', 'MethylLactate'),
        P=0.2 * 101325,
        use_cache=True,
        algorithms=('inside out',)
    )
    esterification.simulate()
    
def test_esterification_column_no_reaction():
    import biosteam as bst
    bst.settings.set_thermo(
        ['Water', 'LacticAcid', 'MethylLactate', 'Methanol', 'SuccinicAcid'],
        pkg='ideal gas',
        cache=True
    )
    
    feed = bst.Stream(
        'feed',
        LacticAcid=4.174,
        Water=5.460,
        SuccinicAcid=0.531,
        MethylLactate=1e-6,
        total_flow=10.165,
        P=0.106 * 101325,
        T=72.5 + 273.15
    )
    feed.T = feed.bubble_point_at_P(P=0.2 * 101325).T
    methanol = bst.Stream('methanol', Methanol=16.695, P=0.2 * 101325, phase='g')
    methanol.T = methanol.dew_point_at_P().T
    esterification = bst.MESHDistillation(
        'esterification',
        ins=(feed, methanol), 
        outs=('esterification_distillate', 'bottoms'),
        N_stages=21,
        feed_stages=(1, 19),
        reflux=1, # Unknown (update later)
        boilup=1,
        # bottoms_product_to_feed=0.0334, # Fraction of feed.
        full_condenser=True,
        maxiter=50,
        LHK=('Methanol', 'MethylLactate'),
        P=0.2 * 101325,
        use_cache=True,
        algorithms=('inside out',)
    )
    esterification.simulate()
    
   
# def test_lactic_acid_ethanol_reactive_distillation():
#     import biosteam as bst
#     from thermosteam.constants import R
#     from math import exp
#     from numpy.testing import assert_allclose
#     bst.settings.set_thermo(['EthylLactate', 'LacticAcid', 'H2O', 'Ethanol'], cache=True)
    
#     class Esterification(bst.KineticReaction):
        
#         def volume(self, stream):
#             return 0.01 # Kg of catalyst
        
#         def rate(self, stream):
#             T = stream.T
#             if T > 365: return 0 # Prevents multiple steady states.
#             kf = 6.52e3 * exp(-4.8e4 / (R * T))
#             kr = 2.72e3 * exp(-4.8e4 / (R * T))
#             LaEt, La, H2O, EtOH = stream.mol / stream.F_mol
#             return 3600 * (kf * La * EtOH - kr * LaEt * H2O) # kmol / kg-catalyst / hr
    
#     rxn = Esterification('LacticAcid + Ethanol -> H2O + EthylLactate', reactant='LacticAcid')
#     stream = bst.Stream(
#         H2O=10, Ethanol=10, LacticAcid=2, T=355,
#     )
#     stream.vle(V=0, P=101325, liquid_conversion=rxn)
#     distillation = bst.MESHDistillation(
#         N_stages=5,
#         ins=[stream],
#         feed_stages=[2],
#         outs=['distillate', 'bottoms_product'],
#         full_condenser=False,
#         reflux=1.0,
#         boilup=2.0,
#         use_cache=True,
#         LHK=('H2O', 'EthylLactate'),
#         stage_reactions={i: rxn for i in range(1,3)},
#         method='fixed-point',
#         maxiter=100,
#     )
#     distillation.simulate()
#     flows = [
#         [6.779963298101173e-06, 3.6172767647177665e-09, 3.1106893353339236, 8.015420206297735],
#         [0.01955718389933403, 1.9804360325200911, 6.908874628528711, 1.9650158298396327],
#     ]
#     for i, j in zip(distillation.outs, flows):    
#         assert_allclose(i.mol, j, rtol=1e-5, atol=1e-3)
#     sys = bst.System.from_units(units=[distillation])
#     sys._setup()
#     sys.run_phenomena()
#     for i, j in zip(distillation.outs, flows):    
#         assert_allclose(i.mol, j, rtol=1e-5, atol=1e-3)
    
# def test_acetic_acid_reactive_distillation():
#     import biosteam as bst
#     import thermosteam as tmo
#     from thermosteam.constants import R
#     import numpy as np
#     from math import exp
#     from numpy.testing import assert_allclose
#     bst.settings.set_thermo(['Water', 'AceticAcid', 'MethylAcetate', 'Methanol'], cache=True)
#     Gamma = tmo.equilibrium.DortmundActivityCoefficients(bst.settings.chemicals)
#     class Esterification(bst.KineticReaction):
#         def volume(self, stream): # kg of catalyst
#             rho_cat = 770 # kg / m3
#             liquid_volume = self.liquid_volume
#             catalyst_volume = 0.5 * liquid_volume
#             catalyst_mass = catalyst_volume * rho_cat
#             return catalyst_mass
        
#         def rate(self, stream): # kmol/kg-catalyst/hr
#             F_mol = stream.F_mol
#             if not F_mol: return 0
#             K_AcOH = 3.15
#             K_MeOH = 5.64
#             K_MeOAc = 4.15
#             K_H2O = 5.24
#             K_adsorption = np.array([K_H2O, K_AcOH, K_MeOAc, K_MeOH])
#             MWs = stream.chemicals.MW
#             k0f = 8.497e6
#             EAf = 60.47
#             k0r = 6.127e5
#             EAr = 63.73
#             T = stream.T
#             mol = stream.mol
#             x = mol / F_mol
#             gamma = Gamma(x, T)
#             activity = gamma * x
#             a = K_adsorption * activity / MWs
#             H2O, LA, BuLA, BuOH = activity
#             kf = k0f * exp(-EAf / (R * T))
#             kr = k0r * exp(-EAr / (R * T))
#             a_sum = sum(a)
#             return 3600 * (kf * LA * BuOH - kr * BuLA * H2O) / (a_sum * a_sum) # kmol / kg-catalyst / hr
        
#     methanol = bst.Stream(
#         Methanol=30,
#     )
#     methanol.vle(V=1, P=101325)
#     acetic_acid = bst.Stream(
#         AceticAcid=10, Water=200
#     )
#     acetic_acid.vle(V=0, P=101325)
#     stage_reactions = {
#         i: Esterification('AceticAcid + Methanol -> MethylAcetate + Water', reactant='AceticAcid')
#         for i in range(3, 9)
#     }
#     distillation = bst.MESHDistillation(
#         N_stages=10,
#         ins=[methanol, acetic_acid],
#         feed_stages=[3, 8],
#         outs=['distillate', 'bottoms_product'],
#         full_condenser=False,
#         reflux=1.0,
#         boilup=2.0,
#         use_cache=True,
#         LHK=('MethylAcetate', 'Water'),
#         stage_reactions=stage_reactions,
#         method='fixed-point',
#         maxiter=100,
#     )
#     distillation.simulate()
#     flows = [
#         [6.585111268569887e-06, 3.6465771244802687e-09,
#          3.110437134095077, 8.015523262914513],
#         [0.019007540635984595, 1.9809858706061696,
#          6.908576991652181, 1.9654626113382356],
#     ]
#     for i, j in zip(distillation.outs, flows):    
#         assert_allclose(i.mol, j, rtol=1e-5, atol=1e-3)
#     sys = bst.System.from_units(units=[distillation])
#     sys._setup()
#     sys.run_phenomena()
#     for i, j in zip(distillation.outs, flows):    
#         assert_allclose(i.mol, j, rtol=1e-5, atol=1e-3)
    
    
# def test_lactic_acid_butanol_reactive_distillation():
#     import biosteam as bst
#     from thermosteam.constants import R
#     from math import exp
#     from numpy.testing import assert_allclose
#     bst.settings.set_thermo(['Water', 'LacticAcid', 'ButylLactate', 'Butanol'], cache=True)
    
#     class Esterification(bst.KineticReaction):
        
#         def volume(self, stream): # kg of catalyst
#             rho_cat = 770 # kg / m3
#             liquid_volume = self.liquid_volume
#             catalyst_volume = 0.5 * liquid_volume
#             catalyst_mass = catalyst_volume * rho_cat
#             return catalyst_mass
        
#         def rate(self, stream): # kmol/kg-catalyst/hr
#             T = stream.T
#             # if T > 370: return 0 # Prevents multiple steady states
#             kf = 2.59e4 * exp(-5.340e4 / (R * T))
#             kr = 3.80e3 * exp(-5.224e4 / (R * T))
#             H2O, LA, BuLA, BuOH = stream.mol / stream.F_mol
#             return 3600 * (kf * LA * BuOH - kr * BuLA * H2O) # kmol / kg-catalyst / hr
        
#     stream = bst.Stream(
#         H2O=10, Butanol=10, LacticAcid=2, T=355,
#     )
#     stream.vle(V=0, P=101325)
#     stage_reactions = {
#         i: Esterification('LacticAcid + Butanol -> Water + ButylLactate', reactant='LacticAcid')
#         for i in range(2)
#     }
#     distillation = bst.MESHDistillation(
#         N_stages=5,
#         ins=[stream],
#         feed_stages=[2],
#         outs=['distillate', 'bottoms_product'],
#         full_condenser=False,
#         reflux=1.0,
#         boilup=2.0,
#         use_cache=True,
#         LHK=('Butanol', 'ButylLactate'),
#         stage_reactions=stage_reactions,
#         method='fixed-point',
#         maxiter=100,
#     )
#     distillation.simulate()
#     flows = [
#         [6.585111268569887e-06, 3.6465771244802687e-09,
#          3.110437134095077, 8.015523262914513],
#         [0.019007540635984595, 1.9809858706061696,
#          6.908576991652181, 1.9654626113382356],
#     ]
#     for i, j in zip(distillation.outs, flows):    
#         assert_allclose(i.mol, j, rtol=1e-5, atol=1e-3)
#     sys = bst.System.from_units(units=[distillation])
#     sys._setup()
#     sys.run_phenomena()
#     for i, j in zip(distillation.outs, flows):    
#         assert_allclose(i.mol, j, rtol=1e-5, atol=1e-3)
    
def test_distillation():
    import biosteam as bst
    bst.settings.set_thermo(
        ['Water', 'AceticAcid', 'EthylAcetate'],
        cache=True
    )
    hot_extract = bst.MultiStream(
        phases=('g', 'l'), T=358.05, P=101325,
        g=[('Water', 20.29), 
           ('AceticAcid', 3.872), 
           ('EthylAcetate', 105.2)],
        l=[('Water', 1.878), 
           ('AceticAcid', 0.6224), 
           ('EthylAcetate', 4.311)]
    )
    distillation = bst.MESHDistillation(
        N_stages=10,
        ins=[hot_extract],
        feed_stages=[5],
        outs=['distillate', 'bottoms_product'],
        full_condenser=True,
        reflux=1.0,
        boilup=3.5,
        use_cache=True,
        LHK=('Water', 'AceticAcid'),
    )
    distillation.simulate()
    flows = [
        [22.055922550575424, 0.1438997998268773, 87.15249997005776],
        [0.11207744942457795, 4.350500200173122, 22.35850002994225],
    ]
    for i, j in zip(distillation.outs, flows):    
        assert_allclose(i.mol, j, rtol=1e-3, atol=1e-3)
    
if __name__ == '__main__':
    test_multi_stage_adiabatic_vle()
    test_distillation()
    # test_esterification_column()





# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from .tank import MixTank
from .heat_exchange import HXutility

__all__ = ('EnzymeTreatment',)

class EnzymeTreatment(MixTank):
    """Create an EnzymeTreatment unit that is cost as a MixTank with a heat exchanger."""
    _N_outs = 1
    auxiliary_unit_names = ('heat_exchanger',)
    
    def _init(self, vessel_type='Conventional', tau=1.0, V_wf=0.9,
                 vessel_material='Stainless steel', T=None):
        MixTank._init(self, vessel_type=vessel_type,
                      tau=tau, V_wf=V_wf, vessel_material=vessel_material)
        self.T = T #: Operating temperature (if any)
        self.heat_exchanger = HXutility(None, None, None, T=T, thermo=self.thermo)
    
    def _run(self):
        out = self.outs[0]
        if self.T is None:
            out.mix_from(self.ins)
        else:
            out.mix_from(self.ins, energy_balance=False)
            out.T = self.T
        
    def _design(self):
        super()._design()
        self.heat_exchanger.simulate_as_auxiliary_exchanger(self.ins, self.outs, self.Hnet)


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
try: from chaospy import distributions as shape
except: pass
import biosteam as bst

__all__ = ('load_default_parameters',
           'add_all_cost_item_params',
           'add_flow_rate_param',
           'add_all_stream_price_params',
           'add_stream_price_param',
           'add_power_utility_price_param',
           'add_basic_TEA_params',
           'triang',
           'bounded_triang',)

# %% Predefined shapes

def triang(mid, proportion=0.1, addition=0): # pragma: no cover
    lb = (1.-proportion)*mid - addition
    ub = (1.+proportion)*mid + addition
    if lb > ub: ub, lb = lb, ub
    return shape.Triangle(lb, mid, ub)

def bounded_triang(mid, lb=0, ub=1, proportion=0, addition=0.1): # pragma: no cover
    if lb > ub: ub, lb = lb, ub
    lower = (1.-proportion)*mid - addition
    upper = (1.+proportion)*mid + addition
    if lower > upper: upper, lower = lower, upper
    if lower < lb: lower = lb
    if upper > ub: upper = ub
    return shape.Triangle(lower, mid, upper)

# %% Fill model with parameters

def load_default_parameters(self, feedstock, shape=triang,
                            bounded_shape=bounded_triang,
                            operating_days=False, include_feedstock_price=True): # pragma: no cover
    """
    Load all default parameters, including coefficients of cost items,
    stream prices, electricity price, heat utility prices, feedstock
    flow rate, and number of operating days.
    
    Parameters
    ----------
    feedstock : Stream
        Main feedstock of process.
    shape : function, optional
        Should take in baseline value and return a chaospy.Dist object or
        None. Default function returns a chaospy.Triangle object with bounds
        at +- 10% of baseline value. The distribution is applied to all
        parameters except exponential factor "n" of cost items.
    bounded_shape : function, optional
        Should take in baseline value and return a chaospy.Dist object or
        None. Default function returns a chaospy.Triangle object with bounds
        at +- 0.1 of baseline value or minimum 0 and maximum 1. The 
        distribution is applied to exponential factor "n" of cost items.
    operating_days : bool, optional
        If True, include operating days
    
    """
    bounded_shape = bounded_shape or shape
    add_all_cost_item_params(self, shape, bounded_shape)
    add_all_stream_price_params(self, shape, feedstock, include_feedstock_price)
    add_power_utility_price_param(self, shape)
    add_heat_utility_price_params(self, shape)
    add_feedstock_flow_rate_param(self, feedstock, shape)
    add_basic_TEA_params(self, shape, operating_days)

def add_all_cost_item_params(model, shape, exp_shape): # pragma: no cover
    system = model._system
    costunits = [i for i in system.units if i._design or i._cost]
    
    # Get all cost options (without repetition)
    all_unit_lines = []
    all_cost_items = []
    for i in costunits:
        if hasattr(i, 'cost_items'):
            if any([i.cost_items is j for j in all_cost_items]): continue
            all_unit_lines.append(i.line)    
            all_cost_items.append(i.cost_items)
    
    # Add cost exponents, base cost, and electricity cost as parameters
    # (for each column in each dataframe of the list)
    for cost_items, line in zip(all_cost_items, all_unit_lines):
        for ID, item in cost_items.items():
            _cost(model, ID, item, line, shape)
            _exp(model, ID, item, line, exp_shape)
            _kW(model, ID, item, line, shape)
            
def add_all_stream_price_params(model, shape, feedstock, include_feedstock_price): # pragma: no cover
    # Add stream prices as parameters
    system = model._system
    feeds = system.feeds
    products = system.products
    if not include_feedstock_price:
        feeds = feeds.copy()
        feeds.discard(feedstock)
    for s in feeds: add_stream_price_param(model, s, shape)
    for s in products: add_stream_price_param(model, s, shape)

def set_price(price): # pragma: no cover
    bst.PowerUtility.price = price
        
def add_power_utility_price_param(model, shape): # pragma: no cover
    if bst.PowerUtility.price:
        baseline = bst.PowerUtility.price
        model.parameter(set_price, element='Electricity', units='USD/kWhr',
                        distribution=shape(baseline),
                        baseline=baseline)
        
def add_heat_utility_price_params(model, shape): # pragma: no cover
    agents = (*bst.HeatUtility.cooling_agents,
                    *bst.HeatUtility.heating_agents)
    for agent in agents:
        add_agent_price_params(model, agent.ID, agent, shape)
        
class Setter: # pragma: no cover
    __slots__ = ('obj', 'attr')
    def __init__(self, obj, attr):
        self.obj = obj
        self.attr = attr
    def __call__(self, value):
        setattr(self.obj, self.attr, value)

class FeedstockSetter: # pragma: no cover
    __slots__ = ('feedstock', 'system')
    def __init__(self, feedstock, system):
        self.feedstock = feedstock
        self.system = system
    def __call__(self, value):
        self.system.rescale(self.feedstock, value / self.feedstock.F_mass)
        
def add_agent_price_params(model, name, agent, shape): # pragma: no cover
    if agent.heat_transfer_price:
        baseline = agent.heat_transfer_price
        model.parameter(Setter(agent, 'heat_transfer_price'), element=name, units='USD/kJ',
                        name='Price', distribution=shape(baseline), baseline=baseline)
    elif agent.regeneration_price:
        baseline = agent.regeneration_price
        model.parameter(Setter(agent, 'regeneration_price'),
                        element=name, units='USD/kmol',
                        name='Price', distribution=shape(baseline),
                        baseline=baseline)
        
def add_flow_rate_param(model, feed, shape): # pragma: no cover
    baseline = feed.F_mass
    model.parameter(Setter(feed, 'F_mass'), element=feed, units='kg/hr',
                    distribution=shape(baseline), kind='coupled',
                    name='Flow rate', baseline=baseline)
    
def add_feedstock_flow_rate_param(model, feed, shape): # pragma: no cover
    baseline = feed.F_mass
    model.parameter(FeedstockSetter(feed, model.system), element=feed, units='kg/hr',
                    distribution=shape(baseline), kind='coupled',
                    name='Flow rate', baseline=baseline)
    
def add_basic_TEA_params(model, shape, operating_days): # pragma: no cover
    param = model.parameter
    TEA = model._system.TEA
    
    if operating_days:
        baseline = TEA.operating_days
        param(Setter(TEA, 'operating_days'), element='TEA', 
              distribution=shape(baseline),
              baseline=baseline,
              name='Operating days')
    
    baseline = TEA.income_tax
    param(Setter(TEA, 'income_tax'),
                 element='TEA',
                 distribution=shape(baseline),
                 baseline=baseline,
                 name='Income tax')
        
    if TEA.startup_months:
        baseline = TEA.startup_months
        param(Setter(TEA, 'startup_months'), element='TEA', 
              baseline=baseline,
              distribution=shape(baseline), name='startup_months')
            
    

# %% Parameter creators for cost_items

def add_stream_price_param(model, stream, shape): # pragma: no cover
    mid = stream.price
    if not mid: return
    model.parameter(Setter(stream, 'price'),
                    element=stream, units='USD/kg',
                    distribution=shape(mid),
                    baseline=mid,
                    name='price')

def _cost(model, ID, item, line, shape): # pragma: no cover
    key = 'cost'
    mid = float(item[key])
    if not mid: return None
    distribution = shape(mid)
    name = 'base cost'
    if ID!=line: ID = ID + ' ' + name
    else: ID = name
    _cost_option(model, ID, item, key, line, distribution, 'USD', mid)
    
def _exp(model, ID, item, line, shape): # pragma: no cover
    key = 'n'
    mid = float(item[key])
    if not mid: return None
    distribution = shape(mid)
    name = 'exponent'
    if ID!=line: ID = ID + ' ' + name
    else: ID = name
    _cost_option(model, ID, item, key, line, distribution, None, mid)

class CostItemSetter: # pragma: no cover
    __slots__ = ('item', 'key', 'size')
    def __init__(self, item, key, size=1):
        self.item = item
        self.key = key
        self.size = size
    def __call__(self, value):
        self.item[self.key] = value*self.size

def _kW(model, ID, item, line, shape): # pragma: no cover
    key = 'kW'
    mid = float(item[key])
    if not mid: return None
    units = item.units
    size = item.S
    mid = mid/size
    distribution = shape(mid)
    units = 'kW / ' + units
    name = 'electricity rate'
    if ID!=line: ID = ID + ' ' + name
    else: ID = name
    model.parameter(CostItemSetter(item, key, size), element=line,
                    baseline=mid,
                    units=units, distribution=distribution, name=ID)

def _cost_option(model, ID, item, key, line, distribution, units, mid): # pragma: no cover
    model.parameter(CostItemSetter(item, key), element=line, units=units, 
                    baseline=mid,
                    distribution=distribution, name=ID)
 


# -*- coding: utf-8 -*-
"""
This code tests a mock biorefinery system with integrated carbon capture.
It works as follows:
    
1. Simulate everything except carbon capture to get the emissions from the boiler.
2. Simulate the carbon capture. The emissions may change due to additional heat/power demand.
3. Repeat steps 1 and 2 until the emissions do not change.

"""
import biosteam as bst
from numpy.testing import assert_allclose

def test_integrated_facilities():
    
    class Biorefinery(bst.Unit):
        
        def _cost(self):
            self.add_power_utility(20)
            
    
    class CoHeatAndPower(bst.Facility):
        network_priority = 1
        _N_outs = 1
        @property
        def flue_gas(self):
            return self.outs[0]
        
        def _cost(self):
            power = sum([i.power_utility.consumption for i in self.other_units])
            self.add_power_utility(-power)
            self.flue_gas.imol['CO2'] = power / 10
            
    
    class CarbonCapture(bst.Unit):
        
        @property
        def flue_gas(self):
            return self.ins[0]
        @flue_gas.setter
        def flue_gas(self, flue_gas):
            self.ins[0] = flue_gas
        
        def _cost(self):
            power = self.flue_gas.F_mol * 0.1
            self.add_power_utility(power)
        
    bst.settings.set_thermo(['CO2'], cache=True)
    with bst.System('biorefinery') as biorefinery_sys: 
        B = Biorefinery()
        CHP = CoHeatAndPower()
        
    with bst.System('carbon_capture') as carbon_capture_sys: 
        CC = CarbonCapture()
        CC.flue_gas = CHP.flue_gas
        
    full_system = bst.System('full_system', path=[biorefinery_sys, carbon_capture_sys])
    full_system.set_tolerance(mol=1e-6, rmol=1e-6)
    full_system.simulate()
    
    assert_allclose(CC.flue_gas.mol[0], 2.020202)
    assert_allclose(CHP.power_utility.production, 20.20202)
    
if __name__ == '__main__':
    test_integrated_facilities()


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from ._variable import Variable

__all__ = ('Indicator', 'indicator', 'Metric', 'metric')

class Indicator(Variable):
    """
    Create an Indicator object that serves as an argument for Model objects.
    
    Parameters
    ----------
    name : str
        Name of indicator.
    units : str
        Indicator units of measure.
    getter : function
        Should take no arguments and return the indicator value.
    element : str
        Element corresponding to indicator
    
    """
    __slots__ = ('getter', 'last_value')
    distribution = None
    def __init__(self, name, getter, units=None, element=None):
        if name is None and hasattr(getter, '__name__'): name = getter.__name__
        super().__init__(name, units, element)
        self.getter = getter
        
    def __call__(self):
        self.last_value = self.getter()
        return self.last_value
    
    def get(self):
        """Return value of indicator. This method used cached values."""
        try:
            return self.last_value
        except:
            self.last_value = self.getter()
            return self.last_value
    
    def difference(self):
        """Return the difference between the current indicator value and the last one 
        evaluated by calling this object."""
        return self.getter() - self.last_value
    
    
def indicator(getter=None, name=None, units=None, element='Biorefinery'):
    if not getter: return lambda getter: indicator(getter, name, units, element)
    return Indicator(name, getter, units, element)

metric = indicator
Metric = Indicator


# -*- coding: utf-8 -*-
"""
This module contains methods and correlations to find aeration requirements
for bioreactors.

.. contents:: :local:
   
.. autofunction:: biosteam.units.design_tools.aeration.log_mean_driving_force
.. autofunction:: biosteam.units.design_tools.aeration.Henrys_law_constant
.. autofunction:: biosteam.units.design_tools.aeration.C_L
.. autofunction:: biosteam.units.design_tools.aeration.C_O2_L
.. autofunction:: biosteam.units.design_tools.aeration.kLa_stirred_Riet
.. autofunction:: biosteam.units.design_tools.aeration.kla_stirred_Labik
.. autofunction:: biosteam.units.design_tools.aeration.kLa_stirred_Galaction
.. autofunction:: biosteam.units.design_tools.aeration.kla_bubcol_Deshpande
.. autofunction:: biosteam.units.design_tools.aeration.kla_bubcol_DeJesus
.. autofunction:: biosteam.units.design_tools.aeration.kla_bubcol_Akita_Yoshida
.. autofunction:: biosteam.units.design_tools.aeration.kla_bubcol_Posarac_Tekic
.. autofunction:: biosteam.units.design_tools.aeration.kla_bubcol_Seno
.. autofunction:: biosteam.units.design_tools.aeration.kla_bubcol_Suh
.. autofunction:: biosteam.units.design_tools.aeration.kla_bubcol_Shah
.. autofunction:: biosteam.units.design_tools.aeration.kla_bubcol_Dewes
.. autofunction:: biosteam.units.design_tools.aeration.P_at_kLa_Riet

References
----------
.. [1] Van’t Riet, K. Review of Measuring Methods and Results in Nonviscous 
    Gas-Liquid Mass Transfer in Stirred Vessels. Ind. Eng. Chem. Proc. Des. 
    Dev. 1979, 18 (3), 357–364. https://doi.org/10.1021/i260071a001.
.. [2] Deshpande, S. S., Kar, K., Pressler, J., Tebeka, I., Martins, B., Rosenfeld, D., & Biggs, J. (2019).
    Mass transfer estimation for bubble column scale up. Chemical Engineering Science, 205, 350–357.
    https://doi.org/10.1016/j.ces.2019.05.011
.. [3] De Jesus, S. S., Moreira Neto, J., & Maciel Filho, R. (2017). 
    Hydrodynamics and mass transfer in bubble column, conventional airlift,
    stirred airlift and stirred tank bioreactors, using viscous fluid: A comparative study. 
    Biochemical Engineering Journal, 118, 70–81. https://doi.org/10.1016/j.bej.2016.11.019
.. [4] Akita, K., & Yoshida, F. (1973).
    Gas Holdup and Volumetric Mass Transfer Coefficient in Bubble Columns. 
    Effects of Liquid Properties. Industrial & Engineering Chemistry Process Design and Development, 12(1), 76–80. 
    https://doi.org/10.1021/i260045a015
.. [5] Pošarac, D., & Tekić, M. N. (1987). 
    Gas holdup and volumetric mass transfer coefficient in bubble columns with dilute alcohol solutions. 
    AIChE Journal, 33(3), 497–499. https://doi.org/10.1002/aic.690330316
.. [6] Seno, T., Uchida, S., & Tsuyutani, S. (1990).
    Mass transfer in countercurrent and cocurrent bubble columns. Chemical Engineering & Technology, 13(1), 113–118.
    https://doi.org/10.1002/ceat.270130115
.. [7] Suh, I. ‐S., Schumpe, A., Deckwer, W. ‐D., & Kulicke, W. ‐M. (1991). 
    Gas‐liquid mass transfer in the bubble column with viscoelastic liquid. 
    The Canadian Journal of Chemical Engineering, 69(2), 506–512. https://doi.org/10.1002/cjce.5450690215
.. [8] Shah, M., Kiss, A. A., Zondervan, E., Van Der Schaaf, J., & De Haan, A. B. (2012).
    Gas Holdup, Axial Dispersion, and Mass Transfer Studies in Bubble Columns. 
    Industrial & Engineering Chemistry Research, 51(43), 14268–14278. https://doi.org/10.1021/ie301227t
.. [9] Labík, L., Moucha, T., Petříček, R., Rejl, J. F., Valenz, L., & Haidl, J. (2017). 
    Volumetric mass transfer coefficient in viscous liquid in mechanically agitated fermenters. 
    Measurement and correlation. Chemical Engineering Science, 170, 451–463. https://doi.org/10.1016/j.ces.2017.04.006
.. [10] Galaction, A.-I., Cascaval, D., Oniscu, C., & Turnea, M. (2004). 
    Prediction of oxygen mass transfer coefficients in stirred bioreactors for bacteria, yeasts and fungus broths.
    Biochemical Engineering Journal, 20(1), 85–94. https://doi.org/10.1016/j.bej.2004.02.005
.. [11] Benz, G. T. Bioreactor Design for Chemical Engineers. AICHE CEP 2011, 21–26.
.. [12] Dewes, I., & Schumpe, A. (1997). Gas density effect on mass transfer in the slurry bubble column. 
    Chemical Engineering Science, 52(21–22), 4105–4109. https://doi.org/10.1016/S0009-2509(97)00252-2

"""

__all__ = (
    'C_L',
    'C_O2_L',
    'kLa_stirred_Riet',
    'P_at_kLa_Riet',
    'log_mean_driving_force',
)

# from inspect import signature
from math import exp, log, sqrt

#: Henry's law coefficients. Data from NIST Standard Reference Database 69: NIST Chemistry WebBook:
#: https://webbook.nist.gov/cgi/cbook.cgi?ID=C7782447&Mask=10#Notes
H_coefficients = { # k_H, A
    'O2': (0.0013, 1500),
    'CO2': (0.035, 2400),
    'CO': (0.00099, 1300),
    'H2': (0.00078, 500), 
}

def Henrys_law_constant(T, k_H, A):
    """
    Return Henry's law constant [mol / kg / bar] for a chemical given the temperature 
    [K] and the coefficients.
    
    """
    dTinv = (1. / T - 1. / 298.15)
    return k_H * exp(A * dTinv)
    
def C_L(T, Py, chemical):
    """Chemical concentration [mol / kg] in the liquid given the temperature [K] and its
    partial pressure in the gas [bar]."""
    return Py * Henrys_law_constant(T, *H_coefficients[chemical])

def C_O2_L(T, P_O2): # Commonly used, so here for convinience
    """O2 concentration [mol / kg] in the liquid given the temperature [K] and oxygen
    partial pressure of O2 in the gas [bar]."""
    return P_O2 * Henrys_law_constant(T, *H_coefficients['O2'])

kLa_methods = {}
kLa_method_names = {
    'Bubble column': [],
    'Stirred tank': []
}
def register(f):
    # parameters = list(signature(f).parameters)
    name = f.__name__
    _, kind, *authors = name.split('_')
    authors = ' & '.join(authors)
    if kind == 'bubcol': kind = 'Bubble column'
    elif kind == 'stirred': kind = 'Stirred tank'
    else: raise ValueError(f"only 'bubcol' and 'stirred' are valid types; not {kind!r}")
    kLa_methods[kind, authors] = f
    kLa_method_names[kind].append(authors)
    return f

_kLa_coefficients_Riet = {
    # Garcia-Ochoa, F.; Gomez, E. Bioreactor Scale-up and Oxygen Transfer Rate
    # in Microbial Processes: An Overview. Biotechnology Advances 2009, 27 (2),
    # 153–176. https://doi.org/10.1016/j.biotechadv.2008.10.006.
    # Name: (a, b, c) # kLa = a * (P/V) ** b * U**c 
    'Figueiredo & Calderbank': (0.026, 0.6, 0.8),
    "Van't Riet": (0.026, 0.4, 0.5),
}

@register
def kLa_stirred_Riet(P, V, U, coefficients=None):
    """
    Return the lumped mass transfer coefficient and the
    mean bubble specific interfacial area (k_L*a; 1/s) given the
    gassed power input (P; W), the total volume (V; m3), and the
    gas superficial velocity in the reactor (m/s).
    
    Other Parameters
    ----------------
    coefficients : Iterable[float]|str, optional
        Name of author or an iterable in the form of a, b, c. The
        
    Notes
    -----
    The correlation is kLa = a * (P/V) ** b * U ** c.
    """
    if coefficients is None:
        a, b, c = _kLa_coefficients_Riet['Figueiredo & Calderbank']
    elif isinstance(coefficients, str):
        a, b, c = _kLa_coefficients_Riet[coefficients]
    else:
        a, b, c = coefficients
    return a * (P / V) ** b * U ** c

# TODO:
# def get_efficiency_factor():
#     """
#     Return the efficiency factor for mass transfer in
#     bubble column reactors.    
#     Take present that the units of the efficiency factor are [m**-1].
#     This values are 
#         - 0.016-0.028 m**-1 for coarse bubble diffusers
#         - 0.07-0.1 m^-1 for fine bubble diffusers
#     """
#     K = None
#     return K

@register
def kla_bubcol_Deshpande(K, M_l, Rhat, T, rho_l, U_g, coefficients=None):
    """
    Returns the KLa coefficient for a bubble column reactor based on 
    the Deshpande et al. (2019) correlation [2]_.
    
    Parameters
    ----------
    K: float
        Efficiency change per unit change in liquid height above sparger, [m^-1]
    M_l: float
        Molar mass, e.g. water =  0.018 kg/mol, [kg/mol]
    Rhat: float
        Ideal gas constant, [J/mol-K]
    T: float
        Temperature, [K]
    rho_l: float
        Mass density of the liquid, e.g. 1000 kg/m^3 for water [kg/m^3]
    U_g: float
        Superficial gas velocity, [m/s]
    
    """
    H_at_T = Henrys_law_constant(T)
    alpha = K * (H_at_T * M_l)/(Rhat * T * rho_l)
    if coefficients is None: coefficients = (1,)
    beta, = coefficients
    return alpha * U_g ** beta

@register
def kla_bubcol_DeJesus(Q, mu, k, n, system_like = None, coefficients=None):
    """
    Returns the KLa coefficient for a bubble column reactor based on the
    De Jesus et al. (2017) correlation [3]_.
    
    Parameters
    ----------
    Q: float
        Specific air flow rate, [vmm]
    mu: float
        Dynamic viscosity, [Pa.s]
    k: float
        Consistency index, [Pa.s^n]
    n: float
        Flow behavior index
    system_like: str
        Name of the system, e.g. "Glycerol" or "Xanthan"
    
    """
    if system_like == "Glycerol":
        if coefficients is None: coefficients = (1.27e-2, 0.51, -0.12)
        a, c, d = coefficients
        return a * Q ** c * mu ** d
    elif system_like == "Xanthan":
        if coefficients is None: coefficients = (5.80e-3, 0.66, -0.50, -0.80)
        e, h, i, j = coefficients
        return e * Q ** h * k ** i * n ** j
    else:
        if coefficients is None and system_like is None:
            raise ValueError(
                "To use the De Jesus et al. (2017) correlation without\n a "
                "system_like, please provide the necessary coefficients"
            )
        else:
            if system_like == 'Newtonian':
                a, c, d = coefficients
                return a * Q ** c * mu ** d
            elif system_like == 'Non-Newtonian':
                e, h, i, j = coefficients
                return e * Q ** h * k ** i * n ** j
            else:
                raise ValueError(
                    "system_like must be either 'Newtonian' or 'Non-Newtonian'"
                )

@register
def kla_bubcol_Akita_Yoshida(D, mu_l, rho_l, D_l, g, sigma_l, epsilon_g, coefficients=None):
    """
    Returns the KLa coefficient for a bubble column reactor based on the
    Akita & Yoshida (1973) correlation [4]_.
    
    Parameters
    ----------
    D: float
        Diameter of the column, [m]
    mu_l: float
        Viscosity of the liquid, [Cp]
    rho_l: float    
        Density of the liquid, [kg/m^3]
    D_l: float
        Diffusivity ofliquid, [m^2/s]
    g: float
        Gravitational acceleration, [m/s^2] 
    sigma_l: float
        The surface tension ofliquid [N/m]
    epsilon_g: float
        Overall gashold-up
    
    Notes
    -----
    This correlation is valid under the following range of values:
    D = 0.15 m -> Diameter of the column
    H = 4 m -> Liquid height in the column
    V_g = 0 - 0.33 m/s -> Superficial gasvelocity [m/s]
    And the conditions are for:
    ->  O2-Water, Aqueous Solutions of Glycerol, Glycol, Methanol, Sodium Sulphite
    """
    kla = (D_l/(D**2)) * 0.6 *(mu_l/(rho_l * D_l))**(0.5) * ((g*rho_l*D**2)/sigma_l)**(0.62) * ((g * D**3 * rho_l**2)/mu_l)**(0.31) * epsilon_g**(1.1)
    return kla

@register
def kla_bubcol_Posarac_Tekic(D, mu_l, rho_l, D_l, g, sigma_l, epsilon_g, coefficients=None):

    """
    Returns the KLa coefficient for a bubble column reactor based on the 
    Pošarac & Tekić (1987) [5]_.

    Parameters
    ----------
    D: float
        Diameter of the column, [m]
    mu_l: float
        Viscosity of the liquid, [Cp]
    rho_l: float    
        Density of the liquid, [kg/m^3]
    D_l: float
        Diffusivity ofliquid, [m^2/s]
    g: float
        Gravitational acceleration, [m/s^2] 
    sigma_l: float
        The surface tension ofliquid [N/m]
    epsilon_g: float
        Overall gashold-up

    Notes
    -----
    This correlation is valid under the following range of values:
    D = 0.1 m -> Diameter of the column
    H = 2.5 m -> Liquid height in the column
    V_g = 0.008 - 0.08 m/s -> Superficial gas velocity [m/s]
    And the conditions are:
    ->  Air-Water, Aqueous Solution of Methanol, Ethanol, i-Propanol, nButanol
    """
    kla = (D_l/(D**2)) * 0.961 *(mu_l/(rho_l * D_l))**(0.5) * ((g*rho_l*D**2)/sigma_l)**(0.62) * ((g * D**3 * rho_l**2)/mu_l**2)**(0.31) * epsilon_g**(1.1)

    return kla

@register
def kla_bubcol_Seno(D, mu_l, rho_l, D_l, g, sigma_l, epsilon_g, V_g, V_l, coefficients=None):
    """
    Returns the KLa coefficient for a bubble column reactor based on the 
    Seno et al. (1990) correlation [6]_.
    
    Parameters
    ----------
    D: float
        Diameter of the column, [m]
    mu_l: float
        Viscosity of the liquid, [Cp]
    rho_l: float    
        Density of the liquid, [kg/m^3]
    D_l: float
        Diffusivity ofliquid, [m^2/s]
    g: float
        Gravitational acceleration, [m/s^2] 
    sigma_l: float
        The surface tension ofliquid [N/m]
    epsilon_g: float
        Overall gashold-up
    V_g: float
        Superficial gas velocity, [m/s]
    V_l: float
        Superficial liquid velocity, [m/s]
    
    Notes
    -----
    This correlation is valid under the following range of values:
    D = 0.0464 m -> Diameter of the column
    H = 1.36 m -> Liquid height in the column
    V_g = 0.005 - 0.04 m/s -> Superficial gas velocity [m/s]
    V_l = 0.005 - 0.1 m/s -> Superficial liquid velocity [m/s]
    rho_l = 995 - 1043 kg/m^3 -> Density of the liquid [kg/m^3]
    mu_l = 0.653- 1.31 Cp -> Viscosity of the liquid [Cp]
    sigma_l = 0.0348-0.0728 N/m -> The surface tension ofliquid [N/m]
    D_l = 1.68-3.24 × 10^-9 m^2/s
    Conditions:
    O2- Water, Aqueous Solution of Butanol, Polyoxyethylene sorbitan monolaurate with silicone oil

    """
    kla = (D_l/(D**2)) * 0.6 * (V_g/(V_g+V_l))**-0.39 *(mu_l/(rho_l * D_l))**(0.5) * ((g * rho_l**2 * D**2)/sigma_l)**(0.62) * ((g * D**3 * rho_l**2)/mu_l**2)**(0.31) * epsilon_g**(1.1)
    return kla

@register
def kla_bubcol_Suh(D, mu_l, rho_l, D_l, g, sigma_l, epsilon_g, V_g, V_l, coefficients=None):
    """
    Returns the KLa coefficient for a bubble column reactor based on the 
    Suh et al. (1991) correlation [7]_.
    
    Parameters
    ----------
    D: float
        Diameter of the column, [m]
    mu_l: float
        Viscosity of the liquid, [Cp]
    rho_l: float    
        Density of the liquid, [kg/m^3]
    D_l: float
        Diffusivity ofliquid, [m^2/s]
    g: float
        Gravitational acceleration, [m/s^2] 
    sigma_l: float
        The surface tension ofliquid [N/m]
    epsilon_g: float
        Overall gashold-up
    V_g: float
        Superficial gas velocity, [m/s]
    V_l: float
        Superficial liquid velocity, [m/s]

    Notes
    -----
    This correlation is valid under the following range of values:
    D: 0.15 m -> Diameter of the column [m]
    H: 2.9 m -> Liquid height in the column [m]
    V_g: 0.005 - 0.04 m/s -> Superficial gas velocity [m/s]
    rho_l: 1001 - 1264 kg/m^3 -> Density of the liquid [kg/m^3]
    sigma_l: 0.0656-0.0746 N/m -> The surface tension of the liquid [N/m]
    D_l: 0.318-2.5 * 10^-9 m^2/s -> Diffusivity of liquid [m^2/s]

    And the conditions are:
    Air-Aqueous Sucrose Solution
    """

    kla = (D_l/(D**2)) * 0.018 * (mu_l/(rho_l * D_l))**(0.5) * ((g*rho_l*D**2)/sigma_l)**(0.75) * (g*D**3*rho_l**2/(mu_l**2))**(0.39)*(V_g/(sqrt(g*D)))**0.51
    return kla

@register
def kla_bubcol_Dewes(V_g, mu_l, rho_g):
    """
    Returns the KLa coefficient for a bubble column reactor based on the Dewes & Schumpe (1997) correlation [12]_.
    
    Parameters
    ----------
    V_g: float
        Superficial gas velocity, [m/s]
    mu_l: float
        Viscosity of the liquid, [mPa.s]
    rho_g: float
        Density of the gas, [kg/m^3]
    
    Notes
    -----
    This correlation is valid under the following range of values:
    D_c: 0.115 m -> Diameter of the column [m]
    H_l: 1.37 m -> Liquid height in the column [m]
    V_g: 0.01 - 0.08 m/s -> Superficial gas velocity [m/s]
    mu_l: 1.35 - 583 mPa.s -> Viscosity of the liquid [mPa.s]
    rho_g: 0.4 - 18.8 kg/m^3 -> Density of the gas [kg/m^3]
    
    """

    kla = V_g**0.90 * mu_l**(-0.55) * rho_g**0.46
    return kla

@register
def kla_bubcol_Shah(D, V_l, g, V_g, rho_l, mu_l, coefficients=None):
    """
    Returns the KLa coefficient for a bubble column reactor based on the 
    Shah et al. (2012) correlation [8]_.
    
    Parameters
    ----------
    D: float
        Diameter of the column, [m]
    V_l:
        Superficial liquid velocity, [m/s]
    g: 
        Gravitational acceleration, [m/s^2]
    V_g:
        Superficial gas velocity, [m/s]
    rho_l:
        Density of the liquid, [kg/m^3]
    mu_l:   
        Viscosity of the liquid, [Cp]
    
    Notes
    -----
    This correlation is valid under the following range of values:
    D: 0.29 m -> Diameter of the column [m]
    H: 0.2 m -> Liquid height in the column [m]
    V_g: 0.021 - 0.105 m/s -> Superficial gas velocity [m/s]
    V_l: 0.0005 - 0.002 m/s -> Superficial liquid velocity [m/s]
    mu_l: 1 - 50 Cp -> Viscosity of the liquid [Cp]
    """
    kla = (V_l/D) * 0.0029 * (V_g**2 / (g * D))**0.301 * (V_l**2/(g * D))**-0.511 * (g*rho_l**3 * D**3 / (mu_l**2))**0.12
    return kla

@register
def kla_stirred_Labik(N, D, v_s, P0, coefficients=None):
    """
    Returns the KLa coefficient for a stirred tank reactor based on the 
    Labik et al. (2017) correlation [9]_.
    
    Parameters
    ----------
    N: float
        impeller frequency [s^-1]
    D: float
        impeller diameter [m]
    v_s: float
        gas superficial velocity [m/s]
    P0: float
        impeller power number (P_u/rhh N^3 D^5) [-]
    
    """
    kla = 0.295 * (N * D)**2.083 * v_s**0.461 * P0 ** 0.737
    return kla

def _kLa_submerged(alpha, beta, gamma, delta, C_x, P_a, V, v_s):
    """
    Auxiliary function to calculate kla for submerged aeration.
    """
    return alpha * C_x ** beta * (P_a / V) ** gamma * v_s ** delta
def _kLa_surface(alpha, beta, P, V):
    """
    Auxiliary function to calculate kla for surface aeration.
    """
    return 1/exp(alpha) * (P/V) ** beta

@register
def kLa_stirred_Galaction(
        aeration_type: str, V, C_x, v_s, P_a=None, P=None, 
        organism_type=None, coefficients=None,
    ):
    """
    Return the kLa for a stirred tank reactor based on the 
    Galaction et al. (2004) correlation [10]_.
    
    Parameters
    ----------
    aeration_type: str
        Type of aireation, in this case either 'Surface' or 'Submerged'
    P_a: float
        power consumption for mixing of aerated broths, [W]
    P: float
        power consumption for mixing of non-aerated broths, [W]
    V: float
        volume of the medium, [m**3]
    C_x: float
        Biomass concentration, [g/l dry weight]
    v_s: float
        superficial air velocity, [m/s]
    
    """
    if aeration_type == 'Submerged':
        if organism_type == 'P.shermanii':
            if coefficients is None: coefficients = (6.586, -0.282, 0.0286, 0.429)
            alpha, beta, gamma, delta = coefficients
            return _kLa_submerged(alpha, beta, gamma, delta, C_x, P_a, V, v_s)
        elif organism_type == 'S.cerevisiae':
            if coefficients is None: coefficients = (52.44, -0.702, -0.0762, 0.514)
            alpha, beta, gamma, delta = coefficients
            return _kLa_submerged(alpha, beta, gamma, delta, C_x, P_a, V, v_s)
        elif organism_type == 'P.chrysogenum_pellet':
            if coefficients is None: coefficients = (0.193, -0.269, 0.0288, 0.257)
            alpha, beta, gamma, delta = coefficients
            return _kLa_submerged(alpha, beta, gamma, delta, C_x, P_a, V, v_s)
        elif organism_type == 'P.chrysogenum_mycelia':
            if coefficients is None: coefficients = (33.59, -1.012, -0.0463, 0.94)
            alpha, beta, gamma, delta = coefficients
            return _kLa_submerged(alpha, beta, gamma, delta, C_x, P_a, V, v_s)
        else:
            raise ValueError("Please provide a valid organism_type (in this case 'P.shermanii', 'S.cerevisiae', 'P.chrysogenum_pellet' or 'P.chrysogenum_mycelia')")
    elif aeration_type == 'Surface':
        if organism_type == 'P.shermanii':
            alpha = 10.35 + 0.63 *log(C_x)
            beta = 0.694 - 0.069 * log(C_x)
            return _kLa_surface(alpha, beta, P, V)
        elif organism_type == 'S.cerevisiae':
            alpha = 11.64 + 0.307 * log(C_x)
            beta = 1.14 - 0.187 * log(C_x)
            return _kLa_surface(alpha, beta, P, V)
        elif organism_type == 'P.chrysogenum_pellet':
            alpha = 12.77 + 0.227 * log(C_x)
            beta = 0.62 - 0.109 * log(C_x)
            return _kLa_surface(alpha, beta, P, V)
        elif organism_type == 'P.chrysogenum_mycelia':
            alpha = 11.46 + 0.729 * log(C_x)
            beta = 0.093 + 0.057 * log(C_x)
            return _kLa_surface(alpha, beta, P, V)
        else: 
            raise ValueError("Please provide a valid organism_type (in this case 'P.shermanii', 'S.cerevisiae', 'P.chrysogenum_pellet' or 'P.chrysogenum_mycelia')")
    else:
        raise ValueError("Please provide a valid aeration_type (in this case 'Surface' or 'Submerged')")

def P_at_kLa_Riet(kLa, V, U, coefficients=None):
    """
    Return the gassed power input (P; W) given the lumped mass transfer 
    coefficient and the mean bubble specific interfacial area (k_L*a; 1/s),
    the total volume (V; m3), and the gas superficial velocity in the reactor (m/s).
    
    Other Parameters
    ----------------
    coefficients : Iterable[float]|str, optional
        Name of author or an iterable in the form of a, b, c.
        
    Notes
    ------
    The correlation is kLa = a * (P/V) ** b * U ** c from [1]_.
    
    """
    if coefficients is None:
        a, b, c = _kLa_coefficients_Riet['Figueiredo & Calderbank']
    elif isinstance(coefficients, str):
        a, b, c = _kLa_coefficients_Riet[coefficients]
    else:
        a, b, c = coefficients
    return (kLa / (a * U ** c)) ** (1 / b) * V

def log_mean_driving_force(C_sat_out, C_sat_in, C_out, C_in=None):
    """
    Return the driving force for mass transfer. In small vessels (<1 m tall) 
    where both liquid concentration and saturation are almost constant, the simple
    form is adequate [11]_. In tall vessels, the log-mean driving force should be 
    used for more accuracy, since both the local concentration and the saturation
    concentration are different in the top and bottom of a bioreactor [11]_. 
    
    Parameters
    ----------
    C_sat_out : float
        Saturated concentration exiting the bioreactor.
    C_sat_in : float
        Saturated concentration entering the bioreactor.
    C_out : float, optional
        Outlet concentration. 
    C_in : float
        Inlet concentration. Defaults to the outlet concentration, which
        assumes perfect mixing and oxygen uptake rate is the same everywhere.
        
    """
    if C_in is None: C_in = C_out # Assume perfect mixing and oxygen uptake rate is the same everywhere.
    dC_out = C_sat_out - C_out
    dC_in = C_sat_in - C_in
    if dC_out < 1e-9: # This is not theoretically possible, but may be an assumption
        dC_out = 1e-9 # Assume it near zero
    return (dC_out - dC_in) / log(dC_out / dC_in)

if __name__ == "__main__":
    print('hola')

    repeated_names = [
        ('U_g', ('U_G', 'V_G')),

    ]   
    fermentation_variables_bub_col = {
        'Dashpande': {
        'epsilon_g': (0.5, 1.2), 
        'K': (0.016, 0.028), # for coarse bubble diffusers  'K': (0.07, 0.1), # for fine bubble diffusers
        'M_l': 0.018, # kg/mol
        'Rhat': 8.314, # J/mol-K
        'T': 298.15, # K
        'rho_l': 1000, # kg/m^3
        'U_g': (0.01,0.1), # m/s
        },
        'DeJesus': {
            'Q':(0, 1.6), # vvm
            'mu': (0.001, 0.1), # Pa.s
            'k': (0.1, 100), # Pa.s^n
            'n': (0, 2) # Flow behavior index <1 for pseudoplastic, >1 for dilatant, 1 for Newtonian
        },
        'Akita_Yoshida': {
            'D': 0.15, # m
            'H': 4, # m
            'V_g': (0, 0.33), # m/s
            'mu_l': (0.5, 100), # Cp 1 water, 100 xanthan
            'rho_l': (995, 1043), # kg/m^3
            'D_l': (1.68  * 10**-9, 3.24  * 10**-9) , # m^2/s
            'g': 9.81, # m/s^2
            'sigma_l':  (0.0348,0.0728), # N/m
            'epsilon_g': (0.01, 1) # m/s
        },
        'Posarac_Tekic': {
            'D': 0.1, # m
            'H': 2.5, # m
            'V_g': (0.008, 0.08), # m/s
            'mu_l': (0.5, 100), # Cp 1 water, 100 xanthan
            'rho_l': (995, 1043), # kg/m^3
            'D_l': (1.68  * 10**-9, 3.24  * 10**-9), # m^2/s
            'sigma_l':  (0.0348,0.0728), # N/m
            'epsilon_g': (0.01, 1) # m/s    
        },
        'Seno': {
            'D': 0.0464, # m
            'H': 1.36, # m
            'V_g': (0.005, 0.04), # m/s
            'V_l': (0.005, 0.1), # m/s
            'rho_l': (995, 1043), # kg/m^3
            'mu_l': (0.653, 1.31), # Cp 1 water, 100 xanthan
            'D_l': (1.68 * 10**-9, 3.24 * 10**-9), # m^2/s
            'g': 9.81, # m/s^2
            'sigma_l': (0.0348,0.0728), # N/m
            'epsilon_g': (0.01, 1) # m/s
        },
        'Suh': {
            'D': 0.15, # m
            'H': 2.9, # m
            'V_g': (0.005, 0.04), # m/s
            'rho_l': (1001, 1264), # kg/m^3
            'mu_l': (0.653, 1.31), # Cp 1 water, 100 xanthan
            'D_l': (0.318 * 10**-9, 2.5 * 10**-9) , # m^2/s
            'g': 9.81, # m/s^2
            'sigma_l': (0.0656,0.0746), # N/m
            'epsilon_g': (0.01, 1) # m/s
        },
        'Shah': {
            'D': 0.29, # m
            'H': 0.2, # m
            'V_g': (0.021, 0.105), # m/s
            'V_l': (0.0005, 0.002), # m/s
            'rho_l': (1, 50), # kg/m^3
            'mu_l': (1, 50), # Cp 1 water, 100 xanthan
        },
    }

    fermentation_variables_stirred = {
        'Labik':{
            'N': (4.17, 14.17), # s^-1
            'D':(0.1, 0.2), # m
            'v_s': (2.12 * 1e-3, 8.48 * 1e-3) , # m/s
            'P0': (100, 10000),
        },
        'Galaction':{ # look into what is better for the P_a/V and P/V
            'P_a/V' : (30,120), # W/m^3
            'v_s': (0.8 * 1e-3, 5 * 1e-3), # m/s
            'C_x': (5, 150), # g/l dry weight
            'P/V': (10, 1900), # W/m^3

        }
    }



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
.. contents:: :local:

.. autoclass:: biosteam.units.single_phase_reactor.SinglePhaseReactor

References
----------
.. [1] Seider, W. D.; Lewin, D. R.; Seader, J. D.; Widagdo, S.; Gani, R.; 
    Ng, M. K. Product and Process Design Principles. Wiley 2017.

"""
from .stirred_tank_reactor import AbstractStirredTankReactor
    
__all__ = ('SinglePhaseReactor',)

class SinglePhaseReactor(AbstractStirredTankReactor):
    '''    
    Class for a single phase stirred tank reactor, modeled as a pressure vessel with 
    a given aspect ratio and residence time. A pump-heat exchanger recirculation 
    loop is used to satisfy the duty, if any. By default, a turbine agitator is
    also included if the power usage,`kW_per_m3` , is positive. A vacuum 
    system is also automatically added if the operating pressure is at a vacuum. 

    Parameters
    ----------
    tau :
        Residence time [hr].
    T : 
        Operating temperature [K].
    P : 
        Operating pressure [Pa].
    dT_hx_loop : 
        Maximum change in temperature for the heat exchanger loop. Defaults to 5 K.
    V_wf : 
        Fraction of working volume over total volume. Defaults to 0.8.
    V_max :
        Maximum volume of a reactor [m3]. Defaults to 355.
    kW_per_m3: 
        Power usage of agitator. Defaults to 0.985 [kW / m3] converted from 
        5 hp/1000 gal as in [1]_, for liquid–liquid reaction or extraction.
    vessel_material : 
        Vessel material. Defaults to 'Stainless steel 316'.
    vessel_type : 
        Vessel type. Valid options are 'Horizontal' or 'Vertical'. Defaults to 'Vertical'
    batch :
        Whether to use batch operation mode. If False, operation mode is continuous.
        Defaults to `continuous`.
    tau_0 : 
        Cleaning and unloading time (if batch mode). Defaults to 3 hr.
    N_reactors :
        Number of reactors.
    
    Notes
    -----
    The recirculation loop takes into account the required flow rate needed to
    reach the maximum temperature change of the heat exchanger, `dT_hx_loop`. 
    Increasing `dT_hx_loop` decreases the required recirculation flow rate and
    therefore decreases pump costs.
    
    When parallel reactors are required, one recirculation loop (each with a
    pump and heat exchanger) is assumed. Although it is possible to use the
    same recirculation loop for all reactors, this conservative assumption allows
    for each reactor to be operated independently from each other.
    
    The capital cost for agitators are not yet included in 
    
    Examples
    --------
    Simulate the liquefaction of corn starch 
    (the enzymatic conversion of starch to glucose):
        
    >>> import biosteam as bst
    >>> from biorefineries import corn
    >>> chemicals = corn.create_chemicals()
    >>> bst.settings.set_thermo(chemicals)
    >>> feed = bst.Stream('feed',
    ...     phase='l', 
    ...     T=360.15, P=1.555e+06, 
    ...     Water=1.115e+05, Ethanol=0.1534, 
    ...     Ash=647.3, Yeast=16.68, CaO=4.73, 
    ...     TriOlein=1754, H2SO4=8.207, NH3=78.84, 
    ...     Starch=2.83e+04, Fiber=5548, 
    ...     SolubleProtein=1899, InsolubleProtein=2318, 
    ...     units='kg/hr'
    ... )
    >>> reaction = bst.Reaction('Starch + H2O -> Glucose', 'Starch', 0.95)
    >>> R1 = bst.SinglePhaseReactor('R1',
    ...     ins=feed, outs='product', 
    ...     reaction=reaction,
    ...     tau=0.9, V_wf=0.90, V_max=500.,
    ...     kW_per_m3=0.6,
    ... )
    >>> R1.simulate()
    >>> R1.show('cwt')
    SinglePhaseReactor: R1
    ins...
    [0] feed  
        phase: 'l', T: 360.15 K, P: 1.555e+06 Pa
        composition (%): Water     73.3
                         Ethanol   0.000101
                         Ash       0.426
                         Yeast     0.011
                         CaO       0.00311
                         TriOlein  1.15
                         H2SO4     0.0054
                         ...       25.1
                         --------  1.52e+05 kg/hr
    outs...
    [0] product  
        phase: 'l', T: 360.15 K, P: 1.555e+06 Pa
        composition (%): Water     71.4
                         Ethanol   0.000101
                         Glucose   19.6
                         Ash       0.426
                         Yeast     0.011
                         CaO       0.00311
                         TriOlein  1.15
                         ...       7.41
                         --------  1.52e+05 kg/hr
    
    >>> R1.results()
    Single phase reactor                                  Units                   R1
    Electricity         Power                                kW                 85.6
                        Cost                             USD/hr                 6.69
    Low pressure steam  Duty                              kJ/hr             1.83e+06
                        Flow                            kmol/hr                 47.4
                        Cost                             USD/hr                 11.3
    Design              Reactor volume                       m3                  142
                        Residence time                       hr                  0.9
                        Vessel type                                         Vertical
                        Length                               ft                 38.6
                        Diameter                             ft                 12.9
                        Weight                               lb             1.28e+05
                        Wall thickness                       in                  1.6
                        Vessel material                          Stainless steel 316
    Purchase cost       Vertical pressure vessel            USD             5.44e+05
                        Platform and ladders                USD              3.6e+04
                        Heat exchanger - Floating head      USD             2.16e+04
                        Recirculation pump - Pump           USD             6.31e+03
                        Recirculation pump - Motor          USD                  670
                        Agitator - Agitator                 USD             5.77e+04
    Total purchase cost                                     USD             6.66e+05
    Utility cost                                         USD/hr                   18
    
    '''
    _N_ins = _N_outs = 1
    def _init(self, reaction, **kwargs):
        super()._init(**kwargs)
        self.reaction = reaction
        
    def _run(self):
        feed = self.ins[0]
        product = self.outs[0]
        product.copy_like(feed)
        if self.adiabatic:
            self.adiabatic_reaction(product)
        else:
            self.reaction(product)
            if self.T is not None: product.T = self.T
            if self.P is not None: product.P = self.P


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import time
import numpy as np

__all__ = ('Timer',)

class TimerOffset:
    __slots__ = ['_timer', '_start']
    
    def __init__(self, timer):
        self._timer = timer
        
    def __enter__(self):
        self._start = time.perf_counter()
        
    def __exit__(self, type, exception, traceback):
        self._timer._start += (time.perf_counter() - self._start)
        if exception: raise exception


class Timer: # pragma: no coverage
    """Create a Timer class with functions that measure elapsed time."""
    __slots__ = ['ID', 'record', '_start', 'limit']
    
    class TimesUpError(Exception): pass

    def __init__(self, ID=None, limit=None):
        self.ID = ID
        self.limit = limit #: [float] Maximum time.
        self.record = [] #: [list] Elapsed times from measurments
        self._start = None
        

    def offset(self):
        if self._start is None: 
            raise RuntimeError('timer has not been started yet')
        return TimerOffset(self)

    def start(self):
        """Start timer."""
        # Marks the beginning of a time interval
        self._start = time.perf_counter()
        return self._start

    def mock_start(self):
        self._start = 0
        return 0

    def mock_measure(self):
        record = self.record
        if record: record.append(record[-1] + 1)
        else: record.append(1)

    def measure(self, record=True):
        """Record time interval since last start time."""
        # Appends time difference
        try: elapsed_time = self.elapsed_time
        except TypeError:
            if self._start is None:
                raise RuntimeError("Must run 'start' before 'measure'.")    
        if record: self.record.append(elapsed_time)
        if self.limit is not None and elapsed_time > self.limit:
            raise self.TimesUpError('elapsed time greater than time limit')
        return elapsed_time

    @property
    def elapsed_time(self):
        return time.perf_counter() - self._start

    @property
    def mean(self):
        """The mean value of elapsed time."""
        return np.mean(self.record)

    @property
    def standard_deviation(self):
        """The standard deviation value of elapsed time."""
        return np.std(self.record)

    std = standard_deviation

    @property
    def summary(self):
        """Return statistical summary of the elapsed time as a string."""
        return f"{self.mean:.2g} \u00B1 {self.std:.2g} s"

    def __repr__(self):
        return f"{type(self).__name__}({self.ID})"

    def _info(self):
        ID = ': ' + self.ID if self.ID else ''
        record = np.array(self.record)
        return (f"{type(self).__name__}{ID}\n" 
                f" mean: {self.mean:.2g} second\n"
                f" std: {self.std:.2g} second\n"
                f" record: {record}\n" )
    
    def show(self):
        print(self._info())
    
    _ipython_display_ = show
        


# -*- coding: utf-8 -*-
# Bioindustrial-Park: BioSTEAM's Premier Biorefinery Models and Results
# Copyright (C) 2021-2024, Yalin Li <mailto.yalin.li@gmail.com>
#
# This module is under the UIUC open-source license. See
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.

import math
import biosteam as bst
from . import select_pipe, format_str
brake_eff = bst.design_tools.brake_efficiency
motor_eff = bst.design_tools.motor_efficiency

__all__ = ('WWTpump',)

_hp_to_kW = 0.7457 # auom('hp').conversion_factor('kW')
_lb_to_kg = 0.4536 # auom('lb').conversion_factor('kg')
_ft_to_m = 0.3048 # auom('ft').conversion_factor('m')
_ft3_to_gal = 7.4805 # auom('ft3').conversion_factor('gallon')
_m3_to_gal = 264.1721 # auom('m3').conversion_factor('gallon')


# %%

class WWTpump(bst.Unit):
    """
    Generic class for pumps used in wastewater treatment [6]_.

    Parameters
    ----------
    pump_type : str
        The type of the pump that determines the design algorithms to use.
        The following combination is valid:

            - "permeate_cross-flow"
            - "retentate_CSTR"
            - "retentate_AF"
            - "recirculation_CSTR"
            - "recirculation_AF"
            - "lift"
            - "sludge"
            - "chemical"

    Q_mgd : float
        Volumetric flow rate in million gallon per day, [mgd].
        Will use total volumetric flow through the unit if not provided.
    add_inputs : dict
        Additional inputs that will be passed to the corresponding design algorithm.
        Check the document for the design algorithm for the specific input requirements.
    
    """
    _N_ins = 1
    _N_outs = 1

    v = 3 # fluid velocity, [ft/s]
    C = 110 # Hazen- Williams coefficient for stainless steel (SS)

    _default_equipment_lifetime = {'Pump': 15}

    _valid_pump_types = (
        'permeate_cross-flow',
        'retentate_CSTR',
        'retentate_AF',
        'recirculation_CSTR',
        'recirculation_AF',
        'lift',
        'sludge',
        'chemical'
        )


    def _init(self, pump_type, add_inputs, Q_mgd=None):
        self.pump_type = pump_type
        self.Q_mgd = Q_mgd
        self.add_inputs = add_inputs

    def _run(self):
        self.outs[0].copy_like(self.ins[0])


    def _design(self):
        pump_type = format_str(self.pump_type)
        design_func = getattr(self, f'design_{pump_type}')

        D = self.design_results
        pipe, pumps, hdpe = design_func()
        D['Pipe stainless steel [kg]'] = pipe
        #!!! need to consider pump's lifetime in LCA
        D['Pump stainless steel [kg]'] = pumps
        D['Chemical storage HDPE [m3]'] = hdpe


    def _cost(self):
        self.power_utility.rate = self.BHP/self.motor_efficiency * _hp_to_kW


    # Used by other classes
    @staticmethod
    def _batch_adding_pump(obj, IDs, ins_dct, type_dct, inputs_dct):
        for i in IDs:
            if not hasattr(obj, f'{i}_pump'):
                if getattr(obj, 'system', None):                    
                    if not bst.main_flowsheet is obj.system.flowsheet:
                        bst.main_flowsheet.set_flowsheet(obj.system.flowsheet)
                # Add '.' in ID for auxiliary units
                pump = WWTpump(
                    ID=f'.{obj.ID}_{i}',
                    ins=ins_dct[i],
                    pump_type=type_dct[i],
                    add_inputs=inputs_dct[i])
                setattr(obj, f'{i}_pump', pump)


    # Generic algorithms that will be called by all design functions
    def _design_generic(self, Q_mgd, N_pump, L_s, L_d, H_ts, H_p):
        self.Q_mgd, self._H_ts, self._H_p = Q_mgd, H_ts, H_p
        v, C, Q_cfs = self.v, self.C, self.Q_cfs # [ft/s], -, [ft3/s]

        ### Suction side ###
        # Suction pipe (permeate header) dimensions
        OD_s, t_s, ID_s = select_pipe(Q_cfs/N_pump, v) # [in]

        # Suction friction head, [ft]
        self._H_sf = 3.02 * L_s * (v**1.85) * (C**(-1.85)) * ((ID_s/12)**(-1.17))

        ### Discharge side ###
        # Discharge pipe (permeate collector) dimensions
        OD_d, t_d, ID_d = select_pipe(Q_cfs, v)

        # Discharge friction head, [ft]
        self._H_df = 3.02 * L_d * (v**1.85) * (C**(-1.85)) * ((ID_d/12)**(-1.17))

        ### Material usage ###
        # Pipe SS, assume stainless steel, density = 0.29 lbs/in3
        # SS volume for suction, [in3]
        self._N_pump = N_pump
        V_s = N_pump * math.pi/4*((OD_s)**2-(ID_s)**2) * (L_s*12)
        # SS volume for discharge, [in3]
        V_d =math.pi/4*((OD_d)**2-(ID_d)**2) * (L_d*12)
        # Total SS mass, [kg]
        M_SS_pipe = 0.29 * (V_s+V_d) * _lb_to_kg

        # Pump SS (for pumps within 300-1000 gpm)
        # https://www.godwinpumps.com/images/uploads/ProductCatalog_Nov_2011_spread2.pdf
        # assume 50% of the product weight is SS
        M_SS_pump = N_pump * (725*0.5)

        return M_SS_pipe, M_SS_pump


    def design_permeate_cross_flow(self, Q_mgd=None, cas_per_tank=None, D_tank=None,
                                   TMP=None, include_aerobic_filter=False):
        """
        Design pump for the permeate stream of cross-flow membrane configuration.

        Parameters defined through the `add_inputs` argument upon initialization of
        this unit (Q_mgd listed separately) will be used if not provided
        when calling this function.

        Parameters
        ----------
        Q_mgd : float
            Volumetric flow rate in million gallon per day, [mgd].
        cas_per_tank : int
            Number of membrane cassettes per tank.
        D_tank: float
            Depth of the membrane tank, [ft].
        TMP : float
            Transmembrane pressure, [psi].
        include_aerobic_filter : bool
            Whether aerobic filter is included in the reactor design.
        """
        add_inputs = self.add_inputs
        Q_mgd = Q_mgd or self.Q_mgd
        cas_per_tank = cas_per_tank or add_inputs[0]
        D_tank = D_tank or add_inputs[1]
        TMP = TMP or add_inputs[2]
        include_aerobic_filter = include_aerobic_filter or add_inputs[3]

        H_ts_PERM = D_tank if include_aerobic_filter else 0

        M_SS_IR_pipe, M_SS_IR_pump = self._design_generic(
            Q_mgd=Q_mgd,
            N_pump=cas_per_tank,
            L_s=20, # based on a 30-module unit with a total length of 6 m, [ft]
            L_d=10*cas_per_tank, # based on a 30-module unit with a total width of 1.6 m and extra space, [ft]
            H_ts=H_ts_PERM, #  H_ds_PERM (D_tank) - H_ss_PERM (0 or D_tank)
            H_p=TMP*2.31 # TMP in water head, [ft], comment below on 2.31
            )

        # # factor = 2.31 calculated by
        # factor = auom('psi').conversion_factor('Pa') # Pa is kg/m/s2, now in [Pa]
        # factor /= 9.81 # divided by the standard gravity in m/s2, now in [kg/m2]
        # factor /= 1e3 # divided by water's density in kg/m3, now in [m]
        # factor *= auom('m').conversion_factor('ft') # m to ft

        return M_SS_IR_pipe, M_SS_IR_pump, 0


    def design_retentate_CSTR(self, Q_mgd=None, cas_per_tank=None):
        """
        Design pump for the retentate stream of CSTR reactors.

        Parameters defined through the `add_inputs` argument upon initialization of
        this unit (Q_mgd listed separately) will be used if not provided
        when calling this function.

        Parameters
        ----------
        Q_mgd : float
            Volumetric flow rate in million gallon per day, [mgd].
        cas_per_tank : int
            Number of membrane cassettes per tank.
        """
        Q_mgd = Q_mgd or self.Q_mgd
        cas_per_tank = cas_per_tank or self.add_inputs[0]

        M_SS_IR_pipe, M_SS_IR_pump = self._design_generic(
            Q_mgd=Q_mgd,
            N_pump=cas_per_tank,
            L_s=100, # pipe length per module
            L_d=30, # pipe length per module (same as the discharge side of lift pump)
            H_ts=0., # H_ds_IR (D_tank) - H_ss_IR (D_tank)
            H_p=0. # no pressure
            )

        return M_SS_IR_pipe, M_SS_IR_pump, 0


    def design_retentate_AF(self, Q_mgd=None, N_filter=None, D=None):
        """
        Design pump for the retentate stream of AF reactors.

        Parameters defined through the `add_inputs` argument upon initialization of
        this unit (Q_mgd listed separately) will be used if not provided
        when calling this function.

        Parameters
        ----------
        Q_mgd : float
            Volumetric flow rate in million gallon per day, [mgd].
        N_filter : float
            Number of filter tanks.
        D : float
            Depth of the filter tank, [ft].
        """
        add_inputs = self.add_inputs
        Q_mgd = Q_mgd or self.Q_mgd
        N_filter = N_filter or add_inputs[0]
        D = D or add_inputs[1]

        M_SS_IR_pipe, M_SS_IR_pump = self._design_generic(
            Q_mgd=Q_mgd,
            N_pump=N_filter,
            L_s=100, # assumed pipe length per filter, [ft]
            L_d=30, # same as discharge side of lift pumping, [ft]
            H_ts=0., # H_ds_IR (D) - H_ss_IR (D)
            H_p=0. # no pressure
            )

        return M_SS_IR_pipe, M_SS_IR_pump, 0


    def design_recirculation_CSTR(self, Q_mgd=None, L_CSTR=None):
        """
        Design pump for the recirculation stream of CSTR reactors.

        Parameters defined through the `add_inputs` argument upon initialization of
        this unit (Q_mgd listed separately) will be used if not provided
        when calling this function.

        Parameters
        ----------
        Q_mgd : float
            Volumetric flow rate in million gallon per day, [mgd].
        L_CSTR : float
            Length of the CSTR tank, [ft].
        """
        Q_mgd = Q_mgd or self.Q_mgd
        L_CSTR = L_CSTR or self.add_inputs[0]

        M_SS_IR_pipe, M_SS_IR_pump = self._design_generic(
            Q_mgd=Q_mgd,
            N_pump=1,
            L_s=0., # ignore suction side
            L_d=L_CSTR, # pipe length per train
            H_ts=5., # H_ds_IR (5) - H_ss_IR (0)
            H_p=0. # no pressure
            )

        return M_SS_IR_pipe, M_SS_IR_pump, 0


    def design_recirculation_AF(self, Q_mgd=None, N_filter=None, d=None,
                                D=None):
        """
        Design pump for the recirculation stream of AF reactors.

        Parameters defined through the `add_inputs` argument upon initialization of
        this unit (Q_mgd listed separately) will be used if not provided
        when calling this function.

        Parameters
        ----------
        Q_mgd : float
            Volumetric flow rate in million gallon per day, [mgd].
        N_filter : float
            Number of filter tanks.
        d : float
            diameter of the filter tank, [ft].
        D : float
            Depth of the filter tank, [ft].
        """
        add_inputs = self.add_inputs
        Q_mgd = Q_mgd or self.Q_mgd
        N_filter = N_filter or add_inputs[0]
        d = d or add_inputs[1]
        D = D or add_inputs[2]

        M_SS_IR_pipe, M_SS_IR_pump = self._design_generic(
            Q_mgd=Q_mgd,
            N_pump=N_filter,
            L_s=d+D, # pipe length per filter, [ft]
            L_d=30, # same as discharge side of lift pumping, [ft]
            H_ts=0., # H_ds_IR (D) - H_ss_IR (D)
            H_p=0. # no pressure
            )

        return M_SS_IR_pipe, M_SS_IR_pump, 0


    def design_lift(self, Q_mgd=None, N_filter=None, D=None):
        """
        Design pump for the filter tank to lift streams.

        Parameters defined through the `add_inputs` argument upon initialization of
        this unit (Q_mgd listed separately) will be used if not provided
        when calling this function.

        Parameters
        ----------
        Q_mgd : float
            Volumetric flow rate in million gallon per day, [mgd].
        N_filter : float
            Number of filter tanks.
        D : float
            Depth of the filter tank, [ft].
        """
        add_inputs = self.add_inputs
        Q_mgd = Q_mgd or self.Q_mgd
        N_filter = N_filter or add_inputs[0]
        D = D or add_inputs[1]

        M_SS_IR_pipe, M_SS_IR_pump = self._design_generic(
            Q_mgd=Q_mgd,
            N_pump=N_filter,
            L_s=150, # length of suction pipe per filter, [ft]
            L_d=30, # pipe length per filter, [ft]
            H_ts=D, # H_ds_LIFT (D) - H_ss_LIFT (0)
            H_p=0. # no pressure
            )

        return M_SS_IR_pipe, M_SS_IR_pump, 0


    def design_sludge(self, Q_mgd=None):
        """
        Design pump for handling waste sludge.

        Parameters
        ----------
        Q_mgd : float
            Volumetric flow rate in million gallon per day, [mgd].
        """
        Q_mgd = Q_mgd or self.Q_mgd

        M_SS_IR_pipe, M_SS_IR_pump = self._design_generic(
            Q_mgd=Q_mgd,
            N_pump=1,
            L_s=50, # length of suction pipe, [ft]
            L_d=50, # length of discharge pipe, [ft]
            H_ts=0., # H_ds_LIFT (D) - H_ss_LIFT (0)
            H_p=0. # no pressure
            )

        return M_SS_IR_pipe, M_SS_IR_pump, 0


    def design_chemical(self, Q_mgd=None):
        """
        Design pump for membrane cleaning chemicals (NaOCl and citric acid),
        storage containers are included, and are assumed to be cubic in shape
        and made of HDPE.

        Parameters defined through the `add_inputs` argument upon initialization of
        this unit (Q_mgd listed separately) will be used if not provided
        when calling this function.

        Parameters
        ----------
        Q_mgd : float
            Volumetric flow rate in million gallon per day, [mgd].
        """
        if not Q_mgd:
            V_CHEM = self.ins[0].F_vol * 24 * 7 * 2 # for two weeks of storage, [m3]
            Q_CHEM_mgd = self.Q_mgd
        else:
            V_CHEM = (Q_mgd*1e6/_m3_to_gal) * 7 * 2
            Q_CHEM_mgd = Q_mgd

        # HDPE volume, [m3], 0.003 [m] is the thickness of the container
        V_HDPE = 0.003 * (V_CHEM**(1/3))**2*6
        # # Mass of HDPE, [m3], 950 is the density of the HDPE in [kg/m3]
        # M_HDPE = 950 * V_HDPE

        H_ss_CHEM = V_CHEM**(1/3) / _ft_to_m
        # 9'-7" is the water level in membrane trains
        # 18" is the distance from C/L of the pump to the ground
        H_ds_CHEM = 9 + 7/12 - 18/12
        H_ts_CHEM = H_ds_CHEM - H_ss_CHEM

        M_SS_CHEM_pipe, M_SS_CHEM_pump = self._design_generic(
            Q_mgd=Q_CHEM_mgd,
            N_pump=1,
            L_s=0., # no suction pipe
            L_d=30.,
            H_ts=H_ts_CHEM,
            H_p=0. # no pressure
            )

        return M_SS_CHEM_pipe, M_SS_CHEM_pump, V_HDPE


    @property
    def pump_type(self):
        """
        [str] The type of the pump that determines the design algorithms to use.
        Use `valid_pump_type` to see acceptable pump types.
        """
        return self._pump_type
    @pump_type.setter
    def pump_type(self, i):
        i_lower = i.lower()
        i_lower = i_lower.replace('cstr', 'CSTR')
        i_lower = i_lower.replace('af', 'AF')
        if i_lower not in self.valid_pump_types:
            raise ValueError(f'The given `pump_type` "{i}" is not valid, '
                             'check `valid_pump_types` for acceptable pump types.')
        self._pump_type = i_lower

    @property
    def valid_pump_types(self):
        """[tuple] Acceptable pump types."""
        return self._valid_pump_types

    @property
    def N_pump(self):
        """[int] Number of pumps."""
        return self._N_pump

    @property
    def H_sf(self):
        """[float] Suction friction head, [ft]."""
        return self._H_sf

    @property
    def H_df(self):
        """[float] Discharge friction head, [ft]."""
        return self._H_df

    @property
    def H_ts(self):
        """[float] Total static head, [ft]."""
        return self._H_ts

    @property
    def H_p(self):
        """[float] Pressure head, [ft]."""
        return self._H_p

    @property
    def TDH(self):
        """[float] Total dynamic head, [ft]."""
        return self.H_ts+self.H_sf+self.H_df+self.H_p

    @property
    def BHP(self):
        """[float] Brake horsepower, [hp]."""
        return (self.TDH*self.Q_gpm)/3960/self.brake_efficiency

    @property
    def Q_mgd(self):
        """
        [float] Volumetric flow rate in million gallon per day, [mgd].
        Will use total volumetric flow through the unit if not provided.
        """
        if self._Q_mgd:
            return self._Q_mgd
        return self.F_vol_in*_m3_to_gal*24/1e6
    @Q_mgd.setter
    def Q_mgd(self, i):
        self._Q_mgd = i

    @property
    def Q_gpm(self):
        """[float] Volumetric flow rate in gallon per minute, [gpm]."""
        return self.Q_mgd*1e6/24/60

    @property
    def Q_cmd(self):
        """
        [float] Volumetric flow rate in cubic meter per day, [cmd].
        """
        return self.Q_mgd *1e6/_m3_to_gal # [m3/day]

    @property
    def Q_cfs(self):
        """[float] Volumetric flow rate in cubic feet per second, [cfs]."""
        return self.Q_mgd*1e6/24/60/60/_ft3_to_gal

    @property
    def brake_efficiency(self):
        """[float] Brake efficiency."""
        return brake_eff(self.Q_gpm)

    @property
    def motor_efficiency(self):
        """[float] Motor efficiency."""
        return motor_eff(self.BHP)


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2021-2024, Yalin Li <mailto.yalin.li@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""

import pytest, numpy as np, biosteam as bst
from numpy.testing import assert_allclose
from biosteam import TEA, System, Stream, Unit, Chemical, settings
from biorefineries.tea import create_cellulosic_ethanol_tea

def test_depreciation_schedule():    
    correct_arrs = {
		# https://www.irs.gov/pub/irs-pdf/p946.pdf, table A-1
		'MACRS5': np.array([0.2000, 0.3200, 0.1920, 0.1152, 0.1152, 0.0576]),
		# Below from Excel function (SLN, SYD, DB, DDB)
		'SL5'   : np.array([0.2000, 0.2000, 0.2000, 0.2000, 0.2000]),
		'DDB5'  : np.array([0.4000, 0.2400, 0.1440, 0.0864, 0.0518]),
		'SYD5'  : np.array([0.3333, 0.2667, 0.2000, 0.1333, 0.0667]),
    }
    
    empty_sys = System('Empty')
    tea = TEA(
        system=empty_sys, IRR=0.15,
        duration=(2021, 2031),
        depreciation='MACRS7', income_tax=0.35,
        operating_days=200, lang_factor=3,
        construction_schedule=(0.4, 0.6), 
        startup_months=0, startup_FOCfrac=0, startup_VOCfrac=0,
        startup_salesfrac=0, finance_interest=0, finance_years=0, 
        finance_fraction=0, WC_over_FCI=0.05
    )

    for k, v in correct_arrs.items():
        tea.depreciation = k
        assert_allclose(tea._get_depreciation_array(), v, atol=1e-4)
    
    defined = np.array([0.4000, 0.2400, 0.1400, 0.0900, 0.1300])
    tea.depreciation = defined
    assert_allclose(tea._get_depreciation_array(), defined)
    
    with pytest.raises(TypeError):
        tea.depreciation = np.array([0.5, 0.1, 'bad'])
        
    with pytest.raises(ValueError):
        tea.depreciation = 'MACRS12312'
        
    with pytest.raises(ValueError):
        tea.depreciation = 'bad'

def test_cashflow_consistency():
    settings.set_thermo([
        Chemical('Dummy', default=True, phase='s', MW=1, search_db=False)
    ])
    
    cost = Stream(Dummy=9793.983511363867 - 1280.916880939845, price=1) # Includes co-product electricity credits
    ethanol = Stream(flow=[21978.374283953395], price=0.7198608114634679)
    
    class MockCellulosicEthanolBiorefinery(Unit):
        _N_ins = _N_outs = 1
        
        def _run(self): pass
        
        def _cost(self):
            self.baseline_purchase_costs['Biorefinery'] = 85338080.48935215
            
    class OSBL(Unit):
        N_outs = _N_ins = 0
        
        def _run(self): pass
        
        def _cost(self):
            self.baseline_purchase_costs['Biorefinery'] = 122287135.13152598
            
    unit = MockCellulosicEthanolBiorefinery(ins=cost, outs=ethanol)
    osbl = OSBL()
    sys = System.from_units(units=[unit, osbl])
    sys.simulate()
    tea = create_cellulosic_ethanol_tea(sys, OSBL_units=[osbl])
    table = tea.get_cashflow_table()
    assert_allclose(tea.NPV, 48964863.88462368)
    assert_allclose(tea.NPV, table['Cumulative NPV [MM$]'].iloc[-1]*1e6)
    tea.IRR = tea.solve_IRR()
    assert_allclose(tea.NPV, 0, atol=100)
    assert_allclose(tea.IRR, 0.12196475238550361)
    tea.IRR = 0.10
    ethanol.price = tea.solve_price(ethanol)
    assert_allclose(ethanol.price, 0.6837971746118124)
    assert_allclose(tea.NPV, 0, atol=100)

def test_tea_startup_months():   
    cost = bst.decorators.cost
    # Total installed equipment cost to be $1 MM
    bst.CE = CE = bst.design_tools.CEPCI_by_year[2013]
    @cost('Fake scaler', 'Lumped cost', CE=CE, cost=1e6, S=1, n=1, BM=1)
    class LumpedCost(bst.Unit):
        '''Does nothing but adding given costs.'''
        _units = {'Fake scaler': ''}
        
        def _design(self):
            self.design_results['Fake scaler'] = 1
        
    class TEA(bst.TEA):
        def __init__(
                self,
                system,
                FOC_over_installed=0.5, # annual O&M
                DPI_over_installed=(1+1),
                TDC_over_DPI=(1+0.2)*(1+0.4), # 20% non-installed & 40% indirect
                FCI_over_TDC=1,
                **kwargs,
                ):
            self.FOC_over_installed = FOC_over_installed
            self.DPI_over_installed = DPI_over_installed
            self.TDC_over_DPI = TDC_over_DPI
            self.FCI_over_TDC = FCI_over_TDC
            bst.TEA.__init__(self, system, **kwargs)
    
        def _FOC(self, installed_equipment_cost): # fixed operating cost
            return installed_equipment_cost*self.FOC_over_installed
        
        def _DPI(self, installed_equipment_cost): # direct permanent investment
            return installed_equipment_cost*self.DPI_over_installed
    
        def _TDC(self, DPI): # total depreciable cost
            return DPI*self.TDC_over_DPI
    
        def _FCI(self, TDC): # fixed capital investment
            return TDC*self.FCI_over_TDC
        
    bst.settings.set_thermo([bst.Chemical('Water')])
    reactant = bst.Stream('reactant', Water=1, units='kg/hr')
    # Total annual sales to be $2.5 MM
    annual_sales = 2.51e6
    product = bst.Stream('product', Water=1, price=annual_sales/365/24, units='kg/hr')
    
    U101 = LumpedCost('U101', ins=reactant, outs=product)
    sys = bst.System('sys', path=(U101,))
    sys.simulate()
    startup_salesfrac = 0.5
    startup_months = 16
    tea = TEA(
        system=sys,
        IRR=0.1,
        duration=(2013, 2013+8), # 8 years
        income_tax=0.21,
        construction_schedule=(1,),
        depreciation='MACRS7',
        operating_days=365,
        startup_months=startup_months,
        startup_FOCfrac=0.5,
        startup_VOCfrac=0.5,
        startup_salesfrac=startup_salesfrac,
        lang_factor=None,
        WC_over_FCI=0.05,
        finance_interest=0,
        finance_years=0,
        finance_fraction=0,
    )
    # Below test NPV and discounted cashflow calculation
    table = tea.get_cashflow_table()
    sales = table['Sales [MM$]']
    w0 = (startup_months/12) % 1 # fraction of the year operating below full capacity
    values = np.ones(9) * annual_sales 
    values[0] = 0 # Construction
    values[1] = annual_sales * startup_salesfrac
    values[2] = w0 * annual_sales * startup_salesfrac + (1 - w0) * annual_sales
    assert_allclose(sales, values / 1e6, atol=1e-4)
    assert_allclose(tea.NPV, table.iloc[-1,-1]*1e6, atol=1e-4)

def test_tea():   
    cost = bst.decorators.cost
    # Total installed equipment cost to be $1 MM
    bst.CE = CE = bst.design_tools.CEPCI_by_year[2013]
    @cost('Fake scaler', 'Lumped cost', CE=CE, cost=1e6, S=1, n=1, BM=1)
    class LumpedCost(bst.Unit):
        '''Does nothing but adding given costs.'''
        _units = {'Fake scaler': ''}
        
        def _design(self):
            self.design_results['Fake scaler'] = 1
        
    class TEA(bst.TEA):
        def __init__(
                self,
                system,
                FOC_over_installed=0.5, # annual O&M
                DPI_over_installed=(1+1),
                TDC_over_DPI=(1+0.2)*(1+0.4), # 20% non-installed & 40% indirect
                FCI_over_TDC=1,
                **kwargs,
                ):
            self.FOC_over_installed = FOC_over_installed
            self.DPI_over_installed = DPI_over_installed
            self.TDC_over_DPI = TDC_over_DPI
            self.FCI_over_TDC = FCI_over_TDC
            bst.TEA.__init__(self, system, **kwargs)
    
        def _FOC(self, installed_equipment_cost): # fixed operating cost
            return installed_equipment_cost*self.FOC_over_installed
        
        def _DPI(self, installed_equipment_cost): # direct permanent investment
            return installed_equipment_cost*self.DPI_over_installed
    
        def _TDC(self, DPI): # total depreciable cost
            return DPI*self.TDC_over_DPI
    
        def _FCI(self, TDC): # fixed capital investment
            return TDC*self.FCI_over_TDC
        
    bst.settings.set_thermo([bst.Chemical('Water')])
    reactant = bst.Stream('reactant', Water=1, units='kg/hr')
    # Total annual sales to be $2.5 MM
    product = bst.Stream('product', Water=1, price=2.5e6/365/24, units='kg/hr')
    
    U101 = LumpedCost('U101', ins=reactant, outs=product)
    sys = bst.System('sys', path=(U101,))
    sys.simulate()

    tea = TEA(
        system=sys,
        IRR=0.1,
        duration=(2013, 2013+15), # 15 years
        income_tax=0.21+0.1,
        construction_schedule=(1,),
        depreciation='MACRS7',
        operating_days=365,
        startup_months=0,
        startup_FOCfrac=1,
        startup_VOCfrac=1,
        startup_salesfrac=1,
        lang_factor=None,
        WC_over_FCI=0.05,
        finance_interest=0.08,
        finance_years=10,
        finance_fraction =0.6,
        accumulate_interest_during_construction=False,
    )

    # Below test NPV and discounted cashflow calculation
    table = tea.get_cashflow_table()
    data = np.array([
        # Depreciable capital [MM$]
        # Fixed capital investment [MM$]
        # Working capital [MM$]
        # Depreciation [MM$]
        # Loan [MM$]
       [ 3.36      ,  3.36      ,  0.168     ,  0.        ,  2.016     ,
        # Loan interest payment [MM$]
        # Loan payment [MM$]
        # Loan principal [MM$]
        # Annual operating cost (excluding depreciation) [MM$]
        # Sales [MM$]
         0.16128   ,  0.        ,  2.016     ,  0.        ,  0.        ,
        # Tax [MM$]
        # Incentives [MM$]
        # Taxed earnings [MM$]
        # Forwarded losses [MM$]
        # Net earnings [MM$]
         0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
        # Cash flow [MM$]     
        # Discount factor
        # Net present value (NPV) [MM$]
        # Cumulative NPV [MM$]
        -1.67328   ,  1.        , -1.67328   , -1.67328   ],
       [ 0.        ,  0.        ,  0.        ,  0.480144  ,  0.        ,
         0.16128   ,  0.30044345,  1.87683655,  1.68      ,  2.5       ,
         0.01221789,  0.        ,  0.03941255,  0.        ,  0.02719466,
         0.50733866,  0.90909091,  0.46121696, -1.21206304],
       [ 0.        ,  0.        ,  0.        ,  0.822864  ,  0.        ,
         0.15014692,  0.30044345,  1.72654003,  1.68      ,  2.5       ,
         0.        ,  0.        ,  0.        ,  0.        , -0.30330745,
         0.51955655,  0.82644628,  0.42938558, -0.78267746],
       [ 0.        ,  0.        ,  0.        ,  0.587664  ,  0.        ,
         0.1381232 ,  0.30044345,  1.56421978,  1.68      ,  2.5       ,
         0.        ,  0.        ,  0.        , -0.30330745, -0.06810745,
         0.51955655,  0.7513148 ,  0.39035053, -0.39232693],
       [ 0.        ,  0.        ,  0.        ,  0.419664  ,  0.        ,
         0.12513758,  0.30044345,  1.38891391,  1.68      ,  2.5       ,
         0.        ,  0.        ,  0.        , -0.3714149 ,  0.09989255,
         0.51955655,  0.68301346,  0.35486412, -0.03746282],
       [ 0.        ,  0.        ,  0.        ,  0.300048  ,  0.        ,
         0.11111311,  0.30044345,  1.19958358,  1.68      ,  2.5       ,
         0.        ,  0.        ,  0.        , -0.27152235,  0.21950855,
         0.51955655,  0.62092132,  0.32260374,  0.28514093],
       [ 0.        ,  0.        ,  0.        ,  0.299712  ,  0.        ,
         0.09596669,  0.30044345,  0.99510681,  1.68      ,  2.5       ,
         0.05202753,  0.        ,  0.16783075, -0.0520138 ,  0.16781702,
         0.46752902,  0.56447393,  0.26390794,  0.54904887],
       [ 0.        ,  0.        ,  0.        ,  0.300048  ,  0.        ,
         0.07960854,  0.30044345,  0.77427191,  1.68      ,  2.5       ,
         0.06804765,  0.        ,  0.21950855,  0.        ,  0.1514609 ,
         0.4515089 ,  0.51315812,  0.23169546,  0.78074432],
       [ 0.        ,  0.        ,  0.        ,  0.149856  ,  0.        ,
         0.06194175,  0.30044345,  0.53577021,  1.68      ,  2.5       ,
         0.11460717,  0.        ,  0.36970055,  0.        ,  0.25509338,
         0.40494938,  0.46650738,  0.18891187,  0.9696562 ],
       [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
         0.04286162,  0.30044345,  0.27818838,  1.68      ,  2.5       ,
         0.16106253,  0.        ,  0.51955655,  0.        ,  0.35849402,
         0.35849402,  0.42409762,  0.15203646,  1.12169266],
       [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
         0.02225507,  0.30044345,  0.        ,  1.68      ,  2.5       ,
         0.16106253,  0.        ,  0.51955655,  0.        ,  0.35849402,
         0.35849402,  0.38554329,  0.13821496,  1.25990762],
       [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
         0.        ,  0.        ,  0.        ,  1.68      ,  2.5       ,
         0.2542    ,  0.        ,  0.82      ,  0.        ,  0.5658    ,
         0.5658    ,  0.3504939 ,  0.19830945,  1.45821707],
       [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
         0.        ,  0.        ,  0.        ,  1.68      ,  2.5       ,
         0.2542    ,  0.        ,  0.82      ,  0.        ,  0.5658    ,
         0.5658    ,  0.31863082,  0.18028132,  1.63849839],
       [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
         0.        ,  0.        ,  0.        ,  1.68      ,  2.5       ,
         0.2542    ,  0.        ,  0.82      ,  0.        ,  0.5658    ,
         0.5658    ,  0.28966438,  0.16389211,  1.80239049],
       [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
         0.        ,  0.        ,  0.        ,  1.68      ,  2.5       ,
         0.2542    ,  0.        ,  0.82      ,  0.        ,  0.5658    ,
         0.5658    ,  0.26333125,  0.14899282,  1.95138332],
       [ 0.        ,  0.        , -0.168     ,  0.        ,  0.        ,
         0.        ,  0.        ,  0.        ,  1.68      ,  2.5       ,
         0.2542    ,  0.        ,  0.82      ,  0.        ,  0.5658    ,
         0.7338    ,  0.23939205,  0.17566589,  2.1270492 ]])
    
    assert_allclose(table.values, data, atol=1e-4)
    assert_allclose(tea.NPV, table.iloc[-1,-1]*1e6, atol=1e-4)
    total_interest_payment1 = table['Loan payment [MM$]'].sum()
    total_interest_payment2 = (
        table['Loan interest payment [MM$]'].iloc[1:].sum() + # payment during construction (year 0) is equity/cash
            table['Loan principal [MM$]'].iloc[0])
    assert_allclose(total_interest_payment1, total_interest_payment2, atol=1e-4)


if __name__ == '__main__':
    test_depreciation_schedule()
    test_cashflow_consistency()
    test_tea()
    test_tea_startup_months()
